#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Event Generators '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import pandas as pd
import re
import requests
import six
from six.moves import urllib
from .base import ESPObject, attribute
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name
from .utils import xml
from .windows import BaseWindow

MapURL = collections.namedtuple('MapURL', ['url'])
ListURL = collections.namedtuple('ListURL', ['url'])
SetURL = collections.namedtuple('SetURL', ['url'])


DELIMITERS = ' ,;@&'
INNER = '=:%#!'


def get_list_delimiter(value):
    '''
    Determine good list delimiter for value

    Parameters
    ----------
    value : list or tuple or set
        The list values

    Returns
    -------
    string

    '''
    value = ''.join('%s' % x for x in value)
    for char in DELIMITERS:
        if char not in value:
            return char
    raise ValueError('Could not determine a good delimiter character') 


def get_map_delimiters(value):
    '''
    Determine good map delimiter for value

    Parameters
    ----------
    value : dict
        The map values

    Returns
    -------
    (inner-delim, outer-delim)

    '''
    key = ''.join('%s' % x for x in value.keys())
    value = ''.join('%s' % x for x in value.values())
    outer = inner = None
    for char in DELIMITERS:
        if char not in key and char not in value:
            outer = char 
            break
    for char in INNER:
        if char not in key and char not in value:
            inner = char 
            break
    if outer is None or inner is None:
        raise ValueError('Could not determine a good delimiter character')
    return inner, outer


class Resources(collections.MutableMapping):
    ''' Event Generator Resource Manager '''

    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        if isinstance(value, (MapURL, ListURL, SetURL)):
            self._data[key] = value
        elif isinstance(value, (dict, collections.MutableMapping)):
            self._data[key] = dict(value)
        elif isinstance(value, (list, tuple)):
            self._data[key] = list(value)
        elif isinstance(value, set):
            self._data[key] = set(value)
        elif isinstance(value, six.string_types):
            raise ValueError('Use add_map_urls, add_list_urls, or '
                             'add_set_urls to add URL resources')
        else:
            raise TypeError('Unknown type for resource: %s' % value)

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)

    def add_map_urls(self, **kwargs):
        '''
        Add URLs to map resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of URL resources.  The key is the
            resource name.  The value is the URL.

        '''
        for key, value in six.iteritems(kwargs):
            self._data[key] = MapURL(value)

    def add_list_urls(self, **kwargs):
        '''
        Add URLs to a list resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of URL resources.  The key is the
            resource name.  The value is the URL.

        '''
        for key, value in six.iteritems(kwargs):
            self._data[key] = ListURL(value)

    def add_set_urls(self, **kwargs):
        '''
        Add URLs to set resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of URL resources.  The key is the
            resource name.  The value is the URL.

        '''
        for key, value in six.iteritems(kwargs):
            self._data[key] = SetURL(value)


class EventGenerator(ESPObject):
    '''
    Event generator

    Attributes
    ----------
    autogen_key : bool, optional
        Should keys be generated for events?
    data : CSV or URL or DataFrame, optional
        Event data
    insert_only : bool, optional
        Should duplicate keys be removed?
    name : string
        Name of the event generator
    publish_target : URL-string or Window, optional
        The URL to publish the events to
    resources : Resources dict
        The resources for the event generator
    init : dict
        Initialization parameters
    fields : dict
        Event data fields
    exists_opcode : string
        Code to determine upsert or delete for an event
    event_data : string
        Either a URL to the data, or CSV data

    Parameters
    ----------
    name : string, optional
        Name of the event generator
    insert_only : bool, optional
        Should duplicate keys be removed?
    autogen_key : bool, optional
        Should keys be generated for events?
    publish_target : URL-string or Window, optional
        The URL to publish the events to
    data : CSV or URL or DataFrame, optional
        Event data

    '''

    insert_only = attribute('insert-only', dtype='bool')
    autogen_key = attribute('autogen-key', dtype='bool')
    publish_target = attribute('publish-target', dtype='string')
    
    def __init__(self, name=None, insert_only=True, autogen_key=True,
                 publish_target=None, data=None):
        ESPObject.__init__(self)
        self.name = name or gen_name(prefix='eg_')
        self.insert_only = insert_only
        self.autogen_key = autogen_key
        self.publish_target = publish_target
        self.resources = Resources()
        self.init = collections.OrderedDict()
        self.fields = collections.OrderedDict()
        self.exists_opcode = None
        self.event_data = data

    def __str__(self):
        if self.publish_target:
            return '%s(name=%s, publish_target=%s)' % \
                   (type(self).__name__, repr(self.name), repr(self.publish_target))
        return '%s(name=%s)' % (type(self).__name__, repr(self.name))

    def __repr__(self):
        return str(self)

    def add_initializers(self, **kwargs):
        '''
        Add values to initialize when event generator starts

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of initializers.  The key is the field
            name.  The value is the initialization expression.

        '''
        self.init.update(kwargs)

    def add_fields(self, **kwargs):
        '''
        Add fields to generate

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of field definitions.  The key is the
            field name.  The value is the field expression.

        '''
        self.fields.update(kwargs)

    def add_map_resources(self, **kwargs):
        '''
        Add map resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of map resources.  The key is the name
            of the resource.  The value is a dictionary of value mappings.

        '''
        for key, value in six.iteritems(kwargs):
            if not isinstance(value, (dict, collections.MutableMapping)):
                raise TypeError('Map resources must be a dictionary')
            self.resources[key] = dict(value)

    def add_list_resources(self, **kwargs):
        '''
        Add list resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of list resources.  The key is the name
            of the resource.  The value is a list of values.

        '''
        for key, value in six.iteritems(kwargs):
            if not isinstance(value, (list, tuple, set)):
                raise TypeError('List resources must be either a list, tuple, or set')
            self.resources[key] = list(value)

    def add_set_resources(self, **kwargs):
        '''
        Add set resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of set resources.  The key is the name
            of the resource.  The value is a set of values.

        '''
        for key, value in six.iteritems(kwargs):
            if not isinstance(value, (list, tuple, set)):
                raise TypeError('Set resources must be either a list, tuple, or set')
            self.resources[key] = set(value)

    def add_map_url_resources(self, **kwargs):
        '''
        Add map URL resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of map URL resources.  The key is the name
            of the resource.  The value is a URL.

        '''
        for key, value in six.iteritems(kwargs):
            self.resources[key] = MapURL(value)

    def add_list_url_resources(self, **kwargs):
        '''
        Add list URL resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of list URL resources.  The key is the name
            of the resource.  The value is a URL.

        '''
        for key, value in six.iteritems(kwargs):
            self.resources[key] = ListURL(value)

    def add_set_url_resources(self, **kwargs):
        '''
        Add set URL resources

        Parameters
        ----------
        **kwargs : keyword-arguments, optional
            Key / value pairs of set URL resources.  The key is the name
            of the resource.  The value is a URL.

        '''
        for key, value in six.iteritems(kwargs):
            self.resources[key] = SetURL(value)

    @property
    def event_data(self):
        if self._event_data is not None:
            return self._event_data
        return self._event_data_url

    @event_data.setter
    def event_data(self, value):
        if isinstance(value, six.string_types):
            if os.path.isfile(value):
                self._event_data_url = None
                with open(value, 'r') as infile:
                    self._event_data = infile.read()
            elif re.match(r'^(https?|ftp|file):', value) and '\n' not in value:
                self._event_data_url = value
                self._event_data = None
            else:
                self._event_data_url = None
                self._event_data = value

        elif isinstance(value, pd.DataFrame):
            self._event_data_url = None
            self._event_data = value.to_csv(header=False, index=False)

        elif hasattr(value, 'read'):
            self._event_data_url = None
            self._event_data = value.read()

        else:
            self._event_data_url = None
            self._event_data = value

    @property
    def publish_target(self):
        '''
        The URL to publish the events to

        Returns
        -------
        string

        '''
        return self._publish_target

    @publish_target.setter
    def publish_target(self, value):
        '''
        Set the URL for publishing events to

        Parameters
        ----------
        value : string or Window
            If a string, it should be the URL to publish to.
            If a Window, the URL will be derived from that object.

        '''
        if isinstance(value, BaseWindow):
            window = value.fullname.replace('.', '/')

            # Get pubsub port
            port = self._get(urllib.parse.urljoin(self.base_url, 'server'),
                             params=get_params(config=True)
                             ).get('pubsub')
            # Get hostname
            hostname = urllib.parse.urlparse(self.base_url).hostname

            value = 'dfESP://%s:%s/%s' % (hostname, port, window)

        self._publish_target = value

    @property
    def is_running(self):
        ''' Is the event generator current running? '''
        out = self._get(urllib.parse.urljoin(self.base_url,
                                             'eventGenerators/%s' % self.name))
        return 'stop' not in out.attrib['state']

    @property
    def url(self):
        '''
        URL of event generator

        Returns
        -------
        string

        '''
        return urllib.parse.urljoin(self.base_url, 'eventGenerators/%s' % self.name)

    @classmethod
    def from_xml(cls, data, session=None):
        '''
        Create event generator from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML event generator definition
        session : requests.Session, optional
            Session that the event generator is associated with

        Returns
        -------
        :class:`EventGenerator`

        '''
        out = cls()
        out.session = session

        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        out._set_attributes(data.attrib)

        publish_target = data.find('./publish-target')
        if publish_target is not None:
            out.publish_target = publish_target.text

        for item in data.findall('./resources/*'):
            if item.tag == 'list-url':
                out.resources.add_list_urls(**{item.attrib['name']: item.text})
            elif item.tag == 'map-url':
                out.resources.add_map_urls(**{item.attrib['name']: item.text})
            elif item.tag == 'set-url':
                out.resources.add_set_urls(**{item.attrib['name']: item.text})
            elif item.tag == 'list':
                out.resources[item.attrib['name']] = \
                    item.text.split(item.attrib.get('delimiter', ' '))
            elif item.tag == 'map':
                outer = item.attrib.get('outer', ' ')
                inner = item.attrib.get('inner', '=')
                out.resources[item.attrib['name']] = \
                    dict([tuple(x.split(inner, 1)) for x in item.text.split(outer)])
            elif item.tag == 'set':
                out.resources[item.attrib['name']] = \
                    set(item.text.split(item.attrib.get('delimiter', ' ')))
            else:
                raise TypeError('Unknown resource type: %s' % item.tag)

        for item in data.findall('./init/value'):
            out.init[item.attrib['name']] = item.text

        for item in data.findall('./fields/field'):
            out.fields[item.attrib['name']] = item.text

        exists_opcode = data.find('./exists-opcode')
        if exists_opcode is not None:
            out.exists_opcode = exists_opcode

        event_source = data.find('./event-source')
        if event_source is not None:
            event_data_url = event_source.find('./event-data-url')
            if event_data_url is not None:
                out.event_data = event_data_url.text
            event_data = event_source.find('./event-data')
            if event_data is not None:
                out.event_data = event_data.text

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Export the event generator definition to an ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('event-generator',
                           xml.get_attrs(self, exclude=['publish_target',
                                                        'exists_opcode',
                                                        'event_data_url',
                                                        'event_data']))

        if self.publish_target:
            xml.add_elem(out, 'publish-target',
                         text_content=self.publish_target)

        if self._event_data_url is not None:
            data = xml.add_elem(out, 'event-source')
            xml.add_elem(data, 'event-data-url', text_content=self._event_data_url)

            fields = xml.add_elem(out, 'fields')
            xml.add_elem(fields, 'field',
                         attrib=dict(name='dummy-id'), text_content=0)

            return out

        if self._event_data is not None:
            data = xml.add_elem(out, 'event-source')
            if isinstance(self._event_data, pd.DataFrame):
                event_data = self._event_data.to_csv(header=False, index=False)
            else:
                event_data = self._event_data
            xml.add_elem(data, 'event-data', text_content=event_data)

            fields = xml.add_elem(out, 'fields')
            xml.add_elem(fields, 'field',
                         attrib=dict(name='dummy-id'), text_content=0)

            return out

        if self.resources:
            resources = xml.add_elem(out, 'resources')
            for key, value in self.resources.items():

                if isinstance(value, list):
                    delim = get_list_delimiter(value)
                    xml.add_elem(resources, 'list',
                                 attrib=dict(name=key, delimiter=delim),
                                 text_content=' '.join(['%s' % x for x in value]))

                elif isinstance(value, set):
                    delim = get_list_delimiter(value)
                    xml.add_elem(resources, 'set',
                                 attrib=dict(name=key, delimiter=delim),
                                 text_content=' '.join(['%s' % x for x in value]))

                elif isinstance(value, dict):
                    inner, outer = get_map_delimiters(value)
                    xml.add_elem(resources, 'map',
                                 attrib=dict(name=key, outer=outer, inner=inner),
                                 text_content=outer.join(['%s%s%s' % (k, inner, v)
                                                          for k, v in value.items()]))

                elif isinstance(value, ListURL):
                    xml.add_elem(resources, 'list-url',
                                 attrib=dict(name=key),
                                 text_content=value.url)

                elif isinstance(value, SetURL):
                    xml.add_elem(resources, 'set-url',
                                 attrib=dict(name=key),
                                 text_content=value.url)

                elif isinstance(value, MapURL):
                    xml.add_elem(resources, 'map-url',
                                 attrib=dict(name=key),
                                 text_content=value.url)

        if self.init:
            init = xml.add_elem(out, 'init')
            for key, value in self.init.items():
                xml.add_elem(init, 'value',
                             attrib=dict(name=key), text_content=value)

        if self.exists_opcode:
            xml.add_elem(out, 'exists-opcode',
                         text_content=self.exists_opcode)

        if self.fields:
            fields = xml.add_elem(out, 'fields')
            for key, value in self.fields.items():
                xml.add_elem(fields, 'field',
                             attrib=dict(name=key), text_content=value)
        else:
            fields = xml.add_elem(out, 'fields')
            xml.add_elem(fields, 'field',
                         attrib=dict(name='dummy-id'), text_content=0)

        return out

    def to_xml(self, pretty=False):
        '''
        Export the event generator definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the XML include whitespace for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def start(self, events=None, blocksize=None, rate=None, pause=None):
        '''
        Start the generator

        Parameters
        ----------
        events : int, optional
            The number of events to inject
        blocksize : int, optional
            The block size of events to inject
        rate : int, optional
            The maximum number of events per second to inject
        pause : int, optional
            The number of milliseconds to pause betwen each event injection

        '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'eventGenerators/%s/state' % self.name),
                  params=get_params(value='started',
                                    events=events,
                                    blocksize=blocksize,
                                    rate=rate,
                                    pause=pause))

    def stop(self):
        ''' Stop the event generator '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'eventGenerators/%s/state' % self.name),
                  params=get_params(value='stopped'))

    def initialize(self):
        ''' Initialize the event generator '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'eventGenerators/%s/state' % self.name),
                  params=get_params(value='initialized'))

    def delete(self):
        ''' Delete the event generator '''
        self._delete(urllib.parse.urljoin(self.base_url,
                                          'eventGenerators/%s' % self.name))

    def save(self, overwrite=True, name=None):
        '''
        Save the event generator

        Parameters
        ----------
        overwrite : bool, optional
            Save the event generator configuration to the server
        name : string, optional
            Optional name override

        '''
        xml = self.to_xml().encode()
        self._put(urllib.parse.urljoin(self.base_url,
                                       'eventGenerators/%s' % (name or self.name)),
                  params=get_params(overwrite=overwrite),
                  data=xml)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP UVC Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class UVCPublisher(Connector):
    '''
    Publish photos taken by a V4L2 compatible

    Parameters
    ----------
    frame_rate : float, optional
        Specifies the frames per second that the camera streams.
        Must be a double. The default value is 15.
    format_in : string, optional
        Specifies the image format of captured photos.
        The default is jpeg. yuyv, an uncompressed image format,
        is also supported.
    format_out : string, optional
        Specifies the image format that the connector publishes.
        The default is jpeg. yuyv, an uncompressed image format,
        is supported only when format_in is yuyv.
    width : int, optional
        Specifies the height of the photo.
    height : int, optional
        Specifies the width of the photo.
    brightness : string, optional
        Specifies the brightness of the photo.
    gain : string, optional
        Specifies the gain of the photo.
    saturation : string, optional
        Specifies the saturation of the photo.
    contrast : string, optional
        Specifies the contrast of the photo.
    device : string, optional
        Specifies the device name the camera is using on the
        Linux operating system.
    blocking : boolean, optional
        Specifies whether the connector is in blocking mode.
    predelay : int, optional
        Specifies a delay time, in seconds, on starting
        the connector.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`UVCPublisher`

    '''
    connector_key = dict(cls='uvc', type='publish')

    property_defs = dict(
        frame_rate=prop('frame_rate', dtype='float'),
        format_in=prop('format_in', dtype='string'),
        format_out=prop('format_out', dtype='string'),
        width=prop('width', dtype='int'),
        height=prop('height', dtype='int'),
        brightness=prop('brightness', dtype='string'),
        gain=prop('gain', dtype='string'),
        saturation=prop('saturation', dtype='string'),
        contrast=prop('contrast', dtype='string'),
        device=prop('device', dtype='string'),
        blocking=prop('blocking', dtype='boolean'),
        predelay=prop('predelay', dtype='int'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, name=None, is_active=None,
                 frame_rate=None, format_in=None, format_out=None,
                 width=None, height=None, brightness=None,
                 gain=None, saturation=None, contrast=None,
                 device=None, blocking=None, predelay=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'uvc', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        properties = map_properties(cls, properties, delete='type')
        return cls(name=name, is_active=is_active, **properties)



import math
import numpy as np

# ----------------------------------------------------------------
# jmp_score.py
# Helper module with functions supporting the Python language
# scoring code generated by JMP
#
# Required by customer code: Yes
# ----------------------------------------------------------------


# return the index of the max value found in an array
# or -1 if all are missing
def max_array(len, lst):
    maxval = -float('inf')
    maxidx = 0
    count_miss = 0

    for i in range(0, len):
        if is_missing(lst[i]):
            count_miss = count_miss + 1
        elif maxval < lst[i]:
            maxval = lst[i]
            maxidx = i
    return -1 if (count_miss == len) else maxidx


# return the index of the min value found in an array
# or -1 if all are missing
def min_array(len, lst):
    minval = float('inf')
    minidx = 0
    count_miss = 0

    for i in range(0, len):
        if is_missing(lst[i]):
            count_miss = count_miss + 1
        elif minval > lst[i]:
            minval = lst[i]
            minidx = i
    return -1 if (count_miss == len) else minidx


def is_missing(x):
    return math.isnan(x) or math.isinf(x) or x is None


def exp(x):
    try:
        return math.exp(x)
    except OverflowError:
        return float('inf')


def pow(a, b=2):
    try:
        return math.pow(a, b)
    except OverflowError:
        return float('inf')


# Also known as logist or logistic
def squish(x):
    return 1.0 / (1.0 + exp(-x))


def squash(x):
    return 1.0 / (1.0 + exp(x))


# Returns true if the numbers are identical using straight comparison.
# If necessary, replace with a suitable comparison using a value of EPSILON
# appropriate for your domain.
def numeq(x, y):
    return x == y
    # return Math.abs(a - b) < EPSILON;


def vec_diag(M):
    """
    Returns the diagonal elements of the square matrix as a vector.
    """
    return np.diag(M).reshape(M.shape[0], 1)


def vec_quadratic(S, X):
    """
    Evaluates as Vec Diag( X * S * X` ).
    """
    if S.shape[1] == X.shape[0]:
        return vec_diag(np.dot(X.T, np.dot(S, X)))
    return vec_diag(np.dot(X, np.dot(S, X.T)))


def sum(S):
    """
    Return the sum of array elements treating missing (NaN) as zero.

    To match JMP's behavior, check if all elements are missing in
    which case missing is returned.
    """
    if np.all(np.isnan(S)):
        return np.nan
    return np.nansum(S)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
from . import utils as tm


SMTP_SERVER = '@ESP_SMTP_SERVER@'
ESP_EMAIL = '@ESP_EMAIL@'


class TestNotificationWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('notification_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest',pubsub='auto',n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('id*:int32','symbol:string','price:double',
                                       'quant:int64','broker:string'),
                               output_insert_only=True, collapse_updates=True)
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                                type='pub',
                                fstype='csv',
                                fsname=os.path.join(tm.DATA_DIR, 'notification_window.csv'))

        notif_win = esp.NotificationWindow(name='notif_win',
                                           schema=('symbol:string','quant:double',
                                                   'price:double','email:string','broker:string'))
        notif_win.set_smtp_settings(SMTP_SERVER)
        notif_win.set_function_context_functions(broker='string($broker)',
            email='rgxReplace($email,\'esptest@mycompany.com\',\'%s\')' % ESP_EMAIL)

        notif_win.add_email('$email', '$email', 'notificationEmailSchema',
                            deliver=True, from_='esp', to='trader',
                            text='broker=$broker symbol=$symbol at $price.',
                            html='<html><head><title>Page Title</title></head>'
                                 '<body><h1>This is a Heading</h1>'
                                 '<p>This is a paragraph.</p></body></html>',
                            image='file:///tmp/Desert.jpeg', test=True)
        notif_win.add_email('$email', '$email', 'notificationEmailSchema',
                            deliver=True, from_='esp', to='trader',
                            text='broker=$broker symbol=$symbol at $price.',
                            html='<html><head><title>Page Title</title></head>'
                                 '<body><h1>This is a Heading</h1>'
                                 '<p>This is a paragraph.</p></body></html>', test=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(notif_win)

        src.add_target(notif_win)

        self._test_model_file('notification_window.xml', proj)

    def test_smtp_xml(self):
        proj = self._load_project('notification_window.xml')
        xml = proj['cq_01']['notif_win'].smtp.to_xml(pretty=True)
        self.assertEqual(xml, '<smtp host="%s" />' % SMTP_SERVER)


class TestEmailNotification(tm.WindowTestCase):

    # source : notification / notificationEmailSchema

    def test_xml(self):
        self._test_model_file('email_notification.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='contquery')

        notif_win = esp.NotificationWindow(name='notify',
                                           schema=('symbol:string','quant:double',
                                                   'price:double','email:string'))
        notif_win.set_smtp_settings(SMTP_SERVER)
        notif_win.add_email('$email', '$email', 'notificationEmailSchema',
                            deliver=True, from_='esp', to='trader',
                            text='I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?',
                            html=r'''<!DOCTYPE html>
                 <html>
                 <head>
                   <title>Page Title</title>
                   </head>

<body>
  <h1>This is a Heading</h1>
    <p>This is a paragraph.</p>
    </body>

</html>''',
                            image='file:///tmp/Desert.jpeg', test=True)

        proj.add_query(cq)
        cq.add_window(notif_win)

        self._test_model_file('email_notification.xml', proj)

    def test_copy(self):
        self._test_copy('email_notification.xml', 'contquery', 'notify')

    def test_xml(self):
        proj = self._load_project('email_notification.xml')
        xml = proj['contquery']['notify'].email[0].to_xml(pretty=True)
        self.assertEqual(xml.strip(), r'''<email test="true">
  <deliver>1</deliver>
  <email-info>
    <sender>$email</sender>
    <recipients>$email</recipients>
    <subject>notificationEmailSchema</subject>
    <from>esp</from>
    <to>trader</to>
  </email-info>
  <email-contents>
    <text-content name="text_content_0">I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?</text-content>
    <html-content name="html_content_0">&lt;!DOCTYPE html&gt;
                 &lt;html&gt;
                 &lt;head&gt;
                   &lt;title&gt;Page Title&lt;/title&gt;
                   &lt;/head&gt;

&lt;body&gt;
  &lt;h1&gt;This is a Heading&lt;/h1&gt;
    &lt;p&gt;This is a paragraph.&lt;/p&gt;
    &lt;/body&gt;

&lt;/html&gt; </html-content>
    <image-content name="image_content_0">file:///tmp/Desert.jpeg</image-content>
  </email-contents>
</email>''')


class TestSMSNotification(tm.WindowTestCase):

    # source : notification / notificationSMS

    def test_xml(self):
        self._test_model_file('sms_notification.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='contquery')

        notif_win = esp.NotificationWindow(name='notify')
        notif_win.set_smtp_settings(SMTP_SERVER)
        notif_win.add_sms('$email', 'notificationSMS', 'esp', 'txt.att.net', '1234567890',
                          deliver=True,
                          text='I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?',
                          test=True)

        proj.add_query(cq)
        cq.add_window(notif_win)

        self._test_model_file('sms_notification.xml', proj)

    def test_copy(self):
        self._test_copy('sms_notification.xml', 'contquery', 'notify')

    def test_xml(self):
        proj = self._load_project('sms_notification.xml')
        xml = proj['contquery']['notify'].sms[0].to_xml(pretty=True)
        self.assertEqual(xml.strip(), r'''<sms test="true">
  <deliver>1</deliver>
  <sms-info>
    <sender>$email</sender>
    <subject>notificationSMS</subject>
    <from>esp</from>
    <gateway>txt.att.net</gateway>
    <phone>1234567890</phone>
  </sms-info>
  <sms-contents>
    <text-content name="text_content_0">I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?</text-content>
  </sms-contents>
</sms>''')


class TestMMSNotification(tm.WindowTestCase):

    # source : notification / notificationMMS

    def test_xml(self):
        self._test_model_file('mms_notification.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='contquery')

        notif_win = esp.NotificationWindow(name='notify')
        notif_win.set_smtp_settings(SMTP_SERVER)
        notif_win.add_mms('$email', 'notificationMMS', 'esp', 'mms.att.net', '1234567890',
                          deliver=True,
                          text='I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?',
                          image='file:///tmp/d.jpeg',
                          test=True)

        proj.add_query(cq)
        cq.add_window(notif_win)

        self._test_model_file('mms_notification.xml', proj)

    def test_copy(self):
        self._test_copy('mms_notification.xml', 'contquery', 'notify')

    def test_xml(self):
        proj = self._load_project('mms_notification.xml')
        xml = proj['contquery']['notify'].mms[0].to_xml(pretty=True)
        self.assertEqual(xml.strip(), r'''<mms test="true">
  <deliver>1</deliver>
  <mms-info>
    <sender>$email</sender>
    <subject>notificationMMS</subject>
    <from>esp</from>
    <gateway>mms.att.net</gateway>
    <phone>1234567890</phone>
  </mms-info>
  <mms-contents>
    <text-content name="text_content_0">I see you traded $quant shares of $symbol at $price. You must have a lot of money. Can I interest you in a timeshare?</text-content>
    <image-content name="image_content_0">file:///tmp/d.jpeg</image-content>
  </mms-contents>
</mms>''')



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from xml.etree import ElementTree as xml
from .base import Window, attribute
from .features import WindowFeature
from .utils import get_args

class TrackerFeature(WindowFeature):
    def __init__(self):
        pass

    def set(self,method = "iou",score_sigma_low = 0.5,score_sigma_high = 0.3,
            iou_sigma = 0.5,iou_sigma2 = 0.3,iou_sigma_dup = 0.0,
            velocity_vector_frames = 15,max_track_lives = 10,
            min_track_length = 0,track_retention = 0):
        self._method = method
        self._score_sigma_low = score_sigma_low
        self._score_sigma_high = score_sigma_high
        self._iou_sigma = iou_sigma
        self._iou_sigma2 = iou_sigma2
        self._iou_sigma_dup = iou_sigma_dup 
        self._velocity_vector_frames = velocity_vector_frames 
        self._max_track_lives = max_track_lives
        self._min_track_length = min_track_length
        self._track_retention = track_retention

    def _feature_to_element(self):
        e = xml.Element("tracker")
        e.attrib["method"] = str(self._method)
        e.attrib["iou-sigma"] = str(self._iou_sigma)
        e.attrib["iou-sigma2"] = str(self._iou_sigma2)
        e.attrib["iou-sigma-dup"] = str(self._iou_sigma_dup)
        e.attrib["score-sigma-low"] = str(self._score_sigma_low)
        e.attrib["score-sigma-high"] = str(self._score_sigma_high)
        e.attrib["max-track-lives"] = str(self._max_track_lives)
        e.attrib["min-track-length"] = str(self._min_track_length)
        e.attrib["velocity-vector-frames"] = str(self._velocity_vector_frames)
        e.attrib["track-retention"] = str(self._track_retention)
        return(e);


class OutputFeature(WindowFeature):
    def __init__(self):
        pass

    def set(self,mode = "wide",prefix = "Object", tracks = 0,
            velocity_vector = False, newborn_tracks = False,
            scale_x = None, scale_y = None):
        self._mode = mode
        self._prefix = prefix
        self._tracks = tracks
        self._velocity_vector = velocity_vector
        self._newborn_tracks = newborn_tracks
        self._scale_x = scale_x
        self._scale_y = scale_y

    def _feature_to_element(self):
        e = xml.Element("output");
        e.attrib["mode"] = str(self._mode)
        e.attrib["velocity-vector"] = str(self._velocity_vector).lower()
        e.attrib["tracks"] = str(self._tracks)
        e.attrib["prefix"] = str(self._prefix)
        e.attrib["newborn-tracks"] = str(self._velocity_vector).lower()
        if self._scale_x != None:
            e.attrib["scale-x"] = str(self._scale_x)
        if self._scale_y != None:
            e.attrib["scale-y"] = str(self._scale_y)
        return(e)

class InputFeature(WindowFeature):
    def __init__(self):
        pass

    def set_rect(self,count = None,score = None, label = None, x = None,y = None,width = None,height = None):
        self._coordType = "rect"
        self._count = count
        self._score = score
        self._label = label
        self._x = x
        self._y = y
        self._width = width
        self._height = height

    def set_yolo(self,count = None,score = None, label = None, x = None,y = None,width = None,height = None):
        self._coordType = "yolo"
        self._count = count
        self._score = score
        self._label = label
        self._x = x
        self._y = y
        self._width = width
        self._height = height

    def set_coco(self,count = None,score = None, label = None, xMin = None,yMin = None,xMax = None,yMax = None):
        self._coordType = "yolo"
        self._count = count
        self._score = score
        self._label = label
        self._xMin = xMin
        self._yMin = yMin
        self._xMax = xMax
        self._yMax = yMax

    def _feature_to_element(self):
        e = xml.Element("input");
        e.attrib["count"] = str(self._count)
        e.attrib["score"] = str(self._score)
        e.attrib["label"] = str(self._label)
        e.attrib["coord-type"] = str(self._coordType)
        if self._coordType == "rect" or self._coordType == "yolo":
            e.attrib["x"] = str(self._x)
            e.attrib["y"] = str(self._y)
            e.attrib["width"] = str(self._width)
            e.attrib["height"] = str(self._height)
        elif self._coordType == "coco":
            e.attrib["x-min"] = str(self._xMin)
            e.attrib["y-min"] = str(self._yMin)
            e.attrib["x-max"] = str(self._xMax)
            e.attrib["y-max"] = str(self._yMax)
        return(e)

class ObjectTrackerWindow(TrackerFeature,OutputFeature,InputFeature,Window):
    '''
    Object Tracker window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
    description : string, optional
        Description of the window

    Attributes
    ----------
    tracker : Tracker
        The window tracker
    output : Output
        The window output
    input : Input
        The window input

    Returns
    -------
    :class:`ObjectTrackerWindow`

    '''

    window_type = 'object-tracker'

    def __init__(self, name=None, pubsub=None, description=None):
        Window.__init__(self, **get_args(locals()))

    def set_tracker(self,method = "iou",score_sigma_low = 0.5,score_sigma_high = 0.3,
                    iou_sigma = 0.5,iou_sigma2 = 0.3,iou_sigma_dup = 0.0,
                    velocity_vector_frames = 15,max_track_lives = 10,
                    min_track_length = 0,track_retention = 0):
        TrackerFeature.set(self,method,score_sigma_low,score_sigma_high,iou_sigma,iou_sigma2,iou_sigma_dup,velocity_vector_frames,max_track_lives,min_track_length,track_retention)
        '''
        Set the tracker

        Parameters
        ----------
        method : string
            Tracking method
        score_sigma_low : float
            Score low detection threshold (σl)
        score_sigma_high : float
            Score high detection threshold (σh)
        iou_sigma : float
            1st iou threshold (σiou)
        iou_sigma2 : float
            2nd iou threshold (σiou-2)
        iou_sigma_dup : float
            Iou duplicate threshold
        velocity_vector_frames : float
            Number of history frames used to calculate the velocity vector
        max_track_lives : float
            Life duration of tracks without detection
        min_track_length : float
            Minimum track length before allowing a missing frame (tmin)
        track_retention : float
            Number of frames the tracks keep the history of positions in memory
        '''

    def set_output(self,mode = "wide",prefix = "Object", tracks = 0,
                    velocity_vector = False, newborn_tracks = False,
                    scale_x = None, scale_y = None):
        '''
        Set the tracker output

        Parameters
        ----------
        mode : string
            wide: values --> rows, long: values --> fields
        prefix : string
            Prefix name of output fields.
        tracks : integer
            Number of tracks to output
        velocity_vector : boolean
            Do we output the velocity vector coordinates?
        newborn_tracks : boolean
            Whether we output the tracks with length < min-track-length
        scale_x : string
            Rescale factor for x dimension. It can be a double or a fractional value (ex '1920/416').
        scale_y : string
            Rescale factor for y dimension. It can be a double or a fractional value (ex '1920/416').
        '''
        OutputFeature.set(self,mode,prefix,tracks,velocity_vector,newborn_tracks,scale_x,scale_x)
 
    def set_input_rect(self,count = None,score = None,label = None,x = None,y = None,width = None,height = None):
        '''
        Set the tracker input

        Parameters
        ----------
        count : string
            Input object count field name
        score : string
            Input object score field name
        label : string
            Input object label field name
        x : string
            Input object x field name
        y : string
            Input object y field name
        width : string
            Input object width field name
        height : string
            Input object height field name
        '''
        InputFeature.set_rect(self,count,score,label,x,y,width,height)
 
    def set_input_yolo(self,count = None,score = None,label = None,x = None,y = None,width = None,height = None):
        '''
        Set the tracker input

        Parameters
        ----------
        count : string
            Input object count field name
        score : string
            Input object score field name
        label : string
            Input object label field name
        x : string
            Input object x field name
        y : string
            Input object y field name
        width : string
            Input object width field name
        height : string
            Input object height field name
        '''
        InputFeature.set_yolo(self,count,score,label,x,y,width,height)
 
    def set_input_coco(self,count = None,score = None,label = None,xMin = None,yMin = None,xMax = None,yMax = None):
        '''
        Set the tracker input

        Parameters
        ----------
        count : string
            Input object count field name
        score : string
            Input object score field name
        label : string
            Input object label field name
        xMin : string
            Input object x min field name
        yMin : string
            Input object y min field name
        xMax : string
            Input object x max field name
        yMax : string
            Input object y max field name
        '''
        InputFeature.set_yolo(self,count,score,label,xMin,yMin,xMax,yMax)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import os
import six
import esppy
import sys
import unittest
from esppy.metadata import Metadata
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class MetadataSubclass(Metadata):

    def __init__(self, session=None):
        Metadata.__init__(self, session=session)
        self._data = {}

    def __metadata__(self):
        return self._data

    def __setitem__(self, key, value):
        self._data[key] = value 

    def __delitem__(self, key):
        del self._data[key]


class TestMetadata(tm.TestCase):

    def setUp(self):
        self.md = Metadata()
        self.mdsub = MetadataSubclass()

    def test_metadata(self):
        with self.assertRaises(NotImplementedError):
            self.md.__metadata__()

        self.assertTrue(isinstance(self.mdsub.__metadata__(), dict))

    def test_getitem(self):
        with self.assertRaises(NotImplementedError):
            self.md['foo']

        with self.assertRaises(KeyError):
            self.mdsub['foo']

        self.mdsub['foo'] = 'bar'

        self.assertEqual(self.mdsub['foo'], 'bar')

    def test_setitem(self):
        with self.assertRaises(NotImplementedError):
            self.md['foo'] = 'bar'

        self.mdsub['foo'] = 'bar'

        self.assertEqual(self.mdsub['foo'], 'bar')

    def test_delitem(self):
        with self.assertRaises(NotImplementedError):
            del self.md['foo']

        with self.assertRaises(KeyError):
            del self.mdsub['foo']

        self.mdsub['foo'] = 'bar'

        self.assertTrue(self.mdsub['foo'], 'bar')

        del self.mdsub['foo']
        
        with self.assertRaises(KeyError):
            self.mdsub['foo']

    def test_iter(self):
        with self.assertRaises(NotImplementedError):
            for key in self.md:
                pass

        self.mdsub.update(dict(a=1, b=2, c=3))

        for i, key in enumerate(sorted(self.mdsub)):
            if i == 0:
                self.assertEqual(key, 'a')
            elif i == 1: 
                self.assertEqual(key, 'b')
            else:
                self.assertEqual(key, 'c')

    def test_len(self):
        with self.assertRaises(NotImplementedError):
            len(self.md)

        self.mdsub.update(dict(a=1, b=2, c=3))

        self.assertEqual(len(self.mdsub), 3)

    def test_str(self):
        with self.assertRaises(NotImplementedError):
            str(self.md)
        self.assertEqual(str(self.mdsub), '{}')

    def test_repr(self):
        with self.assertRaises(NotImplementedError):
            repr(self.md)
        self.assertEqual(repr(self.mdsub), '{}')


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import esppy
import os
import six
import sys
import unittest
from esppy.utils.keyword import dekeywordify, keywordify
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestKeywordUtils(tm.TestCase):

    def test_dekeywordify(self):
        self.assertEqual(dekeywordify('class'), 'class_')
        self.assertEqual(dekeywordify('for'), 'for_')
        self.assertEqual(dekeywordify('foo'), 'foo')

    def test_keywordify(self):
        self.assertEqual(keywordify('class_'), 'class')
        self.assertEqual(keywordify('for_'), 'for')
        self.assertEqual(keywordify('foo_'), 'foo_')


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import os
import pandas as pd
import six
import esppy
import sys
import unittest
from esppy.config import ESP_ROOT
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestEventGenerators(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")
        for key, value in self.s.get_event_generators().items():
            if 'UnitTest' in key:
                self.s.delete_event_generator(key)

    def test_str(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertRegex(str(eg), "EventGenerator\(name=u?'UnitTestEvtGen', publish_target=u?'dfESP://.+?:\d+/%s'\)" % self.query['w_data'].fullname.replace('.', '/'))

    def test_repr(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertRegex(repr(eg), "EventGenerator\(name=u?'UnitTestEvtGen', publish_target=u?'dfESP://.+?:\d+/%s'\)" % self.query['w_data'].fullname.replace('.', '/'))

    def test_event_data(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'], name='UnitTestEvtGen')

        # Set file object
        eg.event_data = open(data_csv_path)
        self.assertEqual(eg.event_data, open(data_csv_path, 'r').read())

        # Set file path
        eg.event_data = data_csv_path
        self.assertEqual(eg.event_data, open(data_csv_path, 'r').read())

        # Set URL
        eg.event_data = 'file://%s' % data_csv_path
        self.assertEqual(eg.event_data, 'file://%s' % data_csv_path)

        # Set CSV content
        eg.event_data = open(data_csv_path, 'rb').read().decode('utf-8')
        self.assertEqual(eg.event_data, open(data_csv_path, 'rb').read().decode('utf-8'))

        # Set DataFrame
        eg.event_data = pd.read_csv(data_csv_path, header=None)
        self.assertEqual(eg.event_data, open(data_csv_path, 'rb').read().decode('utf-8'))

    def test_publish_target(self):
        eg = self.s.create_event_generator(self.query['w_data'], name='UnitTestEvtGen')
        self.assertRegex(eg.publish_target, r'^dfESP://.+?:\d+/%s$' % self.query['w_data'].fullname.replace('.', '/'))

        # Set target to new window
        eg.publish_target = self.query['w_calculate']
        self.assertRegex(eg.publish_target, r'^dfESP://.+?:\d+/%s$' % self.query['w_calculate'].fullname.replace('.', '/'))

        # Set target to explicit URL
        eg.publish_target = 'dfESP://foo.com:12345/project/contquery/window'
        self.assertEqual(eg.publish_target, 'dfESP://foo.com:12345/project/contquery/window')

    def test_is_running(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertFalse(eg.is_running)

        eg.start()
        self.assertTrue(eg.is_running)

        eg.stop()
        self.assertFalse(eg.is_running)

    def test_url(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertRegex(eg.url, r'^https?://.+?:\d+/%s/eventGenerators/%s' % (ESP_ROOT, eg.name))

    def test_from_xml(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        xml = eg.to_xml()
        neweg = eg.from_xml(xml)
        self.assertEqual(neweg.name, eg.name)
        self.assertEqual(neweg.publish_target, eg.publish_target)

    def test_start(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertFalse(eg.is_running)
        eg.start()
        self.assertTrue(eg.is_running)

    @unittest.skip('Hangs waiting for server')
    def test_stop(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        eg.start()
        self.assertTrue(eg.is_running)
        eg.stop()
        self.assertFalse(eg.is_running)

    def test_initialize(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        # TODO: Not sure how to verify this
        eg.initialize()

    def test_delete(self):
        data_csv_path = os.path.join(DATA_DIR, 'data_sa.csv')
        eg = self.s.create_event_generator(self.query['w_data'],
                                           data='file://%s' % data_csv_path,
                                           name='UnitTestEvtGen')
        self.assertIn('UnitTestEvtGen', self.s.get_event_generators())
        eg.delete()
        self.assertNotIn('UnitTestEvtGen', self.s.get_event_generators())


class TestEventResources(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")
        for key, value in self.s.get_event_generators().items():
            if 'UnitTest' in key:
                self.s.delete_event_generator(key)


    def test_types(self):
        from ..evtgen import MapURL, ListURL, SetURL

        eg = self.s.create_event_generator(self.query['w_data'],
                                           name='UnitTestEvtGen')

        eg.resources['dict'] = dict(a=1, b=2)
        eg.resources['list'] = [10, 20]
        eg.resources['set'] = set([100, 200])

        with self.assertRaises(ValueError):
            eg.resources['map-url'] = 'http://esp-server.com/map.txt'

        with self.assertRaises(TypeError):
            eg.resources['map-url'] = None

        eg.resources['map-url'] = MapURL('http://esp-server.com/map.txt')
        eg.resources['list-url'] = ListURL('http://esp-server.com/list.txt')
        eg.resources['set-url'] = SetURL('http://esp-server.com/set.txt')

        eg.resources.add_map_urls(**{'map-url-2': 'http://esp-server.com/map-2.txt'})
        eg.resources.add_list_urls(**{'list-url-2': 'http://esp-server.com/list-2.txt'})
        eg.resources.add_set_urls(**{'set-url-2': 'http://esp-server.com/set-2.txt'})

        self.assertEqual(eg.resources['dict'], dict(a=1, b=2))
        self.assertEqual(eg.resources['list'], [10, 20])
        self.assertEqual(eg.resources['set'], set([100, 200]))

        self.assertEqual(eg.resources['map-url'].url, 'http://esp-server.com/map.txt')
        self.assertEqual(eg.resources['list-url'].url, 'http://esp-server.com/list.txt')
        self.assertEqual(eg.resources['set-url'].url, 'http://esp-server.com/set.txt')

        self.assertEqual(eg.resources['map-url-2'].url, 'http://esp-server.com/map-2.txt')
        self.assertEqual(eg.resources['list-url-2'].url, 'http://esp-server.com/list-2.txt')
        self.assertEqual(eg.resources['set-url-2'].url, 'http://esp-server.com/set-2.txt')

        xml = eg.to_xml(pretty=True)
        neweg = eg.from_xml(xml)

        self.assertEqual(eg.name, neweg.name)
        self.assertEqual(set(eg.resources.keys()), set(neweg.resources.keys()))

    def test_str(self):
        eg = self.s.create_event_generator(self.query['w_data'],
                                           name='UnitTestEvtGen')
        eg.resources['dict'] = dict(a=1)
        eg.resources['list'] = [10, 20]
        eg.resources['set'] = set([100])

        egstr = str(eg.resources)
        self.assertRegex(egstr, r"'dict': \{'a': 1\}")
        self.assertRegex(egstr, r"'set': (\{100\}|set\(\[100\]\))")
        self.assertRegex(egstr, r"'list': \[10, 20\]")

    def test_repr(self):
        eg = self.s.create_event_generator(self.query['w_data'],
                                           name='UnitTestEvtGen')
        eg.resources['dict'] = dict(a=1)
        eg.resources['list'] = [10, 20]
        eg.resources['set'] = set([100])

        egstr = repr(eg.resources)
        self.assertRegex(egstr, r"'dict': \{'a': 1\}")
        self.assertRegex(egstr, r"'set': (\{100\}|set\(\[100\]\))")
        self.assertRegex(egstr, r"'list': \[10, 20\]")

    def test_delitem(self):
        eg = self.s.create_event_generator(self.query['w_data'],
                                           name='UnitTestEvtGen')
        eg.resources['dict'] = dict(a=1)
        eg.resources['list'] = [10, 20]
        eg.resources['set'] = set([100])

        self.assertEqual(set(eg.resources.keys()), set(['dict', 'list', 'set']))

        del eg.resources['dict']
        del eg.resources['list']

        self.assertEqual(set(eg.resources.keys()), set(['set']))

        del eg.resources['set']

        with self.assertRaises(KeyError):
            del eg.resources['set']

        
if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Continuous Query '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
import xml.etree.ElementTree as ET
from six.moves import urllib
from .base import ESPObject, attribute
from .config import get_option
from .metadata import Metadata
from .windows import BaseWindow, get_window_class
from .templates.template import Template
from .utils import xml
from .utils.rest import get_params
from .utils.data import gen_name
from .utils.events import get_events
from .utils.notebook import scale_svg
from .utils.project import expand_path


class WindowDict(collections.MutableMapping):
    '''
    Dictionary for holding window objects

    Attributes
    ----------
    project : string
        The name of the project
    contquery : string
        The name of the continuous query
    session : requests.Session
        The session for the windows

    Parameters
    ----------
    *args : one-or-more arguments, optional
        Positional arguments to MutableMapping
    **kwargs : keyword arguments, optional
        Keyword arguments to MutableMapping

    '''

    def __init__(self, *args, **kwargs):
        collections.MutableMapping.__init__(self, *args, **kwargs)
        self._data = dict()
        self.project = None
        self.project_handle = None
        self.contquery = None
        self.session = None

    @property
    def session(self):
        '''
        The session for the windows

        Returns
        -------
        string

        '''
        return self._session

    @session.setter
    def session(self, value):
        self._session = value
        for item in self._data.values():
            item.session = self._session

    @property
    def project(self):
        '''
        The project that windows are associated with

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        self._project = getattr(value, 'name', value)
        for item in self._data.values():
            item.project = self._project

    @property
    def contquery(self):
        '''
        The continuous query that windows are associated with

        Returns
        -------
        string

        '''
        return self._contquery

    @contquery.setter
    def contquery(self, value):
        self._contquery = getattr(value, 'name', value)
        for item in self._data.values():
            item.contquery = self._contquery
        if hasattr(value, 'project'):
            self.project = value.project

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        if not isinstance(value, BaseWindow):
            raise TypeError('Only Window objects can be values '
                            'in a ContinuousQuery')
        value._register_to_project(self.project_handle)

        oldname = value.name
        if not value.template:
            value.base_name = key
        value.project = self.project
        value.contquery = self.contquery
        value.session = self.session
        self._data[key] = value

        # Make sure targets get updated with new name
        if oldname != key:
            for window in self._data.values():
                for target in set(window.targets):
                    if target.name == oldname:
                        role = target.role
                        window.targets.discard(target)
                        window.add_target(key, role=role)

    def __delitem__(self, key):
        del self._data[key]
        for window in self._data.values():
            window.delete_target(key)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class TemplateDict(WindowDict):
    '''
    Dictionary for holding template objects

    Attributes
    ----------
    project : string
        The name of the project
    contquery : string
        The name of the continuous query
    session : requests.Session
        The session for the windows

    Parameters
    ----------
    *args : one-or-more arguments, optional
        Positional arguments to MutableMapping
    **kwargs : keyword arguments, optional
        Keyword arguments to MutableMapping

    '''

    def __init__(self, *args, **kwargs):
        WindowDict.__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        if not isinstance(value, Template):
            raise TypeError('Only Template objects are valid values')

        value.name = key
        value.project = self.project
        value.contquery = self.contquery
        value.session = self.session
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]


class ContinuousQuery(ESPObject, collections.MutableMapping):
    '''
    Continuous Query

    Parameters
    ----------
    name : string
        Name of the continuous query
    trace : string, optional
        One or more space-separated window names or IDs
    index_type : string, optional
        A default index type for all windows in the continuous query that
        do not explicitly specify an index type
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    timing_threshold : int, optional
        When a window in the query takes more than value microseconds to
        compute for a given event or event block, a warning message is logged
    include_singletons : bool, optional
        Specify whether to add unattached source windows
    description : string, optional
        Description of the continuous query

    Attributes
    ----------
    project : string or Project
        Name of the project the query is associated with
    windows : dict
        Collection of windows in the continuous query
    metadata : dict
        Metadata dictionary
    url : string
        URL of the continuous query

    Notes
    -----
    All parameters are also available as instance attributes.

    Returns
    -------
    :class:`ContinuousQuery`

    '''
    trace = attribute('trace', dtype='string')
    index_type = attribute('index', dtype='string',
                           values={'rbtree': 'pi_RBTREE', 'hash': 'pi_HASH',
                                   'ln_hash': 'pi_LN_HASH', 'cl_hash': 'pi_CL_HASH',
                                   'fw_hash': 'pi_FW_HASH', 'empty': 'pi_EMPTY'})
    timing_threshold = attribute('timing-threshold', dtype='int')
    include_singletons = attribute('include-singletons', dtype='bool')

    def __init__(self, name=None, trace=None, index_type=None,
                 timing_threshold=None, include_singletons=None, description=None):
        self.windows = WindowDict()
        self.templates = TemplateDict()
        ESPObject.__init__(self, attrs=locals())
        self.project = None
        self.name = name or gen_name(prefix='cq_')
        self.description = description
        self.metadata = {}

    @property
    def session(self):
        '''
        The requests.Session object for the continuous query

        Returns
        -------
        string

        '''
        return ESPObject.session.fget(self)

    @session.setter
    def session(self, value):
        ESPObject.session.fset(self, value)
        self.windows.session = value

    @property
    def name(self):
        '''
        The name of the continuous query

        Returns
        -------
        string

        '''
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.windows.contquery = value
        self.templates.contquery = value

    @property
    def project(self):
        '''
        The name of the project

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        self._project = getattr(value, 'name', value)
        self.windows.project = self._project
        self.templates.project = self._project

    def add_window(self, window):
        '''
        Add a window to the continuous query

        Parameters
        ----------
        window : Window
            The Window object to add

        Returns
        -------
        :class:`Window`

        '''
        if not window.name:
            window.name = gen_name(prefix='w_')

        self.windows[window.name] = window

        return window

    def add_windows(self, *windows):
        '''
        Add one or more windows to the continuous query

        Parameters
        ----------
        windows : one-or-more-Windows
            The Window objects to add

        Returns
        -------
        tuple of :class:`Window`s

        '''
        for item in windows:
            self.add_window(item)
        return windows

    def add_template(self, template):
        '''
        Add a template object

        Parameters
        ----------
        template : Template
        a Template object to add to the project

        Returns
        -------
        :class:`Template`

        '''
        if not template.name:
            template.name = gen_name(prefix='t_')

        self.templates[template.name] = template

        for key, window in sorted(six.iteritems(template.windows)):
            self.add_window(window)
        return template

    def delete_templates(self, *templates):
        '''
        Delete templates and related windows

        Parameters
        ----------
        templates : one-or-more strings or Template objects
            The template to delete

        '''
        for item in templates:
            template_key = getattr(item, 'name', item)
            template = self.templates[template_key]
            self.delete_windows(*six.itervalues(template.windows))
            del self.templates[template_key]

        return self.templates

    delete_template = delete_templates

    # def get_template(self, template):
    #     return

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Returns
        -------
        :class:`ContinuousQuery`

        '''
        out = type(self)()

        out.session = self.session
        out.project = self.project

        for key, value in self._get_attributes(use_xml_values=False).items():
            setattr(out, key, value)

        if deep:
            out.windows = dict([(k, v.copy(deep=True))
                                for k, v in self.windows.items()])
        else:
            out.windows.update(self.windows)

        return out

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo):
        return self.copy(deep=True)

    @property
    def fullname(self):
        return '%s.%s' % (self.project, self.name)

    @property
    def url(self):
        '''
        URL of the continuous query

        Returns
        -------
        string

        '''
        if not self.project:
            raise ValueError('This continuous query is not associated with a project.')
        return urllib.parse.urljoin(self.base_url, '%s/%s/' % (self.project, self.name))

    @classmethod
    def from_xml(cls, data, project=None, session=None):
        '''
        Create continous query from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML continuous query definition
        session : requests.Session, optional
            The session object

        Returns
        -------
        :class:`ContinuousQuery`

        '''
        out = cls()
        out.session = session
        out.project = project

        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        out._set_attributes(data.attrib)

        for desc in data.findall('./description'):
            out.description = desc.text

        for item in data.findall('./windows/*'):
            try:
                wcls = get_window_class(item.tag)
            except KeyError:
                raise TypeError('Unknown window type: %s' % item.tag)

            window = wcls.from_xml(item, session=session)
            out.windows[window.name] = window

        for item in data.findall('./edges/*'):
            for target in re.split(r'\s+', item.attrib.get('target', '').strip()):
                if not target:
                    continue
                for source in re.split(r'\s+', item.attrib.get('source', '').strip()):
                    if not source:
                        continue
                    out.windows[source].add_target(target, role=item.get('role'),
                                                   slot=item.get('slot'))

        for item in data.findall('./metadata/meta'):
            if 'id' in item.attrib.keys():
                out.metadata[item.attrib['id']] = item.text
            elif 'name' in item.attrib.keys():
                out.metadata[item.attrib['name']] = item.text

        return out

    from_element = from_xml

    def to_element(self):
        '''
        Export continuous query definition to ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('contquery', xml.get_attrs(self, exclude='project'))

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        if self.metadata:
            metadata = xml.add_elem(out, 'metadata')
            for key, value in sorted(six.iteritems(self.metadata)):
                xml.add_elem(metadata, 'meta', attrib=dict(id=key),
                             text_content=value)

        windows = xml.add_elem(out, 'windows')

        sources = {}
        if self.windows:
            edges = []
            for name, window in sorted(six.iteritems(self.windows)):
                xml.add_elem(windows, window.to_element(query=self))
                for target in window.targets:
                    sources.setdefault(target.name, []).append(window.name)
                    attrib = dict(source=window.name, target=target.name)
                    if target.role:
                        attrib['role'] = target.role
                    if target.slot:
                        attrib['slot'] = target.slot
                    edges.append((target._index, attrib))
            if edges:
                elem = xml.add_elem(out, 'edges')
                for i, attrib in sorted(edges):
                    xml.add_elem(elem, 'edge', attrib=attrib)

        else:
            xml.add_elem(windows,
                get_window_class('window-source')().to_element(query=self))

        # Replace "inherit" data types with the real data type
        n_inherit = -1
        while True:
            inherit = out.findall('./windows/*/schema/fields/field[@type="inherit"]')
            if len(inherit) == n_inherit:
                break
            n_inherit = len(inherit)
            for window in out.findall('./windows/*'):
                for field in window.findall('./schema/fields/field[@type="inherit"]'):
                    for source in sources[window.attrib['name']]:
                        fname = field.attrib['name']
                        if source not in self.windows:
                            raise ValueError("Could not determine data type of "
                                             "field '%s' on window '%s'" % (fname, source))
                        win = self.windows[source]
                        if hasattr(win, 'schema') and fname in win.schema:
                            dtype = win.schema[fname].type
                            field.set('type', dtype)

        return out

    def to_xml(self, pretty=False):
        '''
        Export continuous query definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output embed whitespaced for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def _persist_metadata(self):
        if self.metadata:
            self._set_metadata(self.metadata)

    def _clear_metadata(self):
        self.metadata.clear()

    def _set_metadata(self, data):
        for key, value in six.iteritems(data):
            self._put(urllib.parse.urljoin(self.base_url,
                                           'projectMetadata/%s/%s/%s' %
                                           (self.project, self.name, key)),
                      data='%s' % value)

    def _del_metadata(self, *data):
        for key in data:
            self._delete(urllib.parse.urljoin(self.base_url,
                                              'projectMetadata/%s/%s/%s' %
                                              (self.project, self.name, key)))

    def save_xml(self, dest, mode='w', pretty=True, **kwargs):
        '''
        Save the continuous query XML to a file

        Parameters
        ----------
        dest : string or file-like
            The destination of the XML content
        mode : string, optional
            The write mode for the output file (only used if `dest` is a string)
        pretty : boolean, optional
            Should the XML include whitespace for readability?

        '''
        if isinstance(dest, six.string_types):
            with open(dest, mode=mode, **kwargs) as output:
                output.write(self.to_xml(pretty=pretty))
        else:
            dest.write(self.to_xml(pretty=pretty))

    def to_graph(self, graph=None, schema=False, template_detail=False):
        '''
        Export continuous query definition to graphviz.Digraph

        Parameters
        ----------
        graph : graphviz.Graph, optional
            The parent graph to add to
        schema : bool, optional
            Should window schemas be included?
        template_detail : bool, optional
            Should template detail be shown?

        Returns
        -------
        :class:`graphviz.Digraph`

        '''
        try:
            import graphviz as gv
        except ImportError:
            raise ImportError('The graphviz module is required for exporting to graphs.')

        if graph is None:
            graph = gv.Digraph(format='svg')
            graph.attr('node', shape='rect')
            graph.attr('graph', rankdir='LR', center='false')
            graph.attr('edge', fontname='times-italic')

        if self.windows:
            qgraph = gv.Digraph(format='svg', name='cluster_%s' % self.fullname.replace('.', '_'))
            qgraph.attr('node', shape='rect')
            qgraph.attr('graph', fontname='helvetica')
            qgraph.attr('edge', fontname='times-italic')
            qgraph.attr(label=self.name, labeljust='l',
                        style='filled,rounded', color='#a0a0a0', fillcolor='#f0f0f0',
                        fontcolor='black')

            if self.templates:
                for tkey, template in sorted(self.templates.items()):
                    qgraph.subgraph(template.to_graph(schema=schema, detail=template_detail))

            for wkey, win in sorted(self.windows.items()):
                if not win.template:
                    win.to_graph(graph=qgraph, schema=schema or get_option('display.show_schema'))
                for target in win.targets:
                    if target.name not in self.windows:
                        continue
                    tail_name = win.template.fullname if win.template and not template_detail else win.fullname
                    head_name = target.template.fullname if target.template and not template_detail \
                        else self.windows[target.name].fullname
                    if win.template and target.template and win.template == target.template:
                        continue
                    graph.edge(tail_name, head_name, label=target.role or '')

            graph.subgraph(qgraph)

        else:
            graph.node(self.fullname, label=self.name, labeljust='l',
                       style='filled,rounded', color='#a0a0a0', fillcolor='#f0f0f0',
                       fontcolor='black')

        return graph

    def _repr_svg_(self):
        try:
            return scale_svg(self.to_graph()._repr_svg_())
        except ImportError:
            raise AttributeError('_repr_svg_')

    def __str__(self):
        return '%s(name=%s, project=%s)' % (type(self).__name__,
                                            repr(self.name),
                                            repr(self.project))

    def __repr__(self):
        return str(self)

    def rename_window(self, window, newname):
        '''
        Rename a window and update targets

        Parameters
        ----------
        window : string or Window object
            The window to rename
        newname : string
            The new name of the window

        '''
        self.windows[newname] = self.windows[getattr(window, 'name', window)]
        self.delete_window(window)

    def delete_windows(self, *windows):
        '''
        Delete windows and update targets

        Parameters
        ----------
        windows : one-or-more strings or Window objects
            The window to delete

        '''
        for item in windows:
            del self.windows[getattr(item, 'name', item)]

    delete_window = delete_windows

    def get_windows(self, name=None, type=None, filter=None):
        '''
        Retrieve windows from the server

        Parameters
        ----------
        name : string or list-of-strings, optional
            Names of the windows which you want to retrieve
        type : string or list-of-strings, optional
            Types of windows you want to retrieve
        filter : string or list-of-strings, optional
            Function filter indicating which windows to retrieve

        Notes
        -----
        This method retrieves window definitions from the server, not
        from the collection of windows held locally in this projects
        variables.

        Returns
        -------
        dict of :class:`Window`

        '''
        path = [None] + expand_path(name)
        res = self._get(urllib.parse.urljoin(self.base_url, 'windowXml'),
                        params=get_params(project=self.project,
                                          contquery=self.name,
                                          name=path[-1], type=type,
                                          filter=filter))
        windows = dict()
        for item in res.findall('./*'):
            try:
                wcls = get_window_class(item.tag)
            except KeyError:
                raise TypeError('Unknown window type: %s' % item.tag)
            window = wcls.from_xml(item, session=self.session)
            windows[window.fullname.split('.', 1)[-1]] = window
        return windows

    def get_window(self, name):
        '''
        Retrieve specified window

        Parameters
        ----------
        name : string
            Name of the window

        Notes
        -----
        This method retrieves window definitions from the server, not
        from the collection of windows held locally in this projects
        variables.

        Returns
        -------
        :class:`Window`

        '''
        out = self.get_windows(name)
        if out:
            return list(out.values())[0]
        raise KeyError("No window with the name '%s'." % name)

#
# MutableMapping methods
#

    def __getitem__(self, key):
        return self.windows[key]

    def __setitem__(self, key, value):
        self.windows[key] = value

    def __delitem__(self, key):
        del self.windows[key]

    def __iter__(self):
        return iter(self.windows)

    def __len__(self):
        return len(self.windows)

    def __contains__(self, value):
        return value in self.windows



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Utilities for testing windows '''

from __future__ import print_function, division, absolute_import, unicode_literals

import copy
import os
import re
import sys
import unittest
import six
import esppy
from .. import utils as tm
from ..utils import xml

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = os.path.normpath(os.path.join(tm.get_data_dir(), '..', 'windows', 'data'))


class WindowTestCase(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def _load_project(self, model):
        model_file = os.path.join(DATA_DIR, model)
        model_xml = tm.normalize_xml(open(model_file, 'r').read())
        return self.s.Project.from_xml(model_xml)

    def _test_model_file(self, model, project=None):
        model_file = os.path.join(DATA_DIR, model)
        model_xml = tm.normalize_xml(open(model_file, 'r').read())

        if project is None:
            proj1 = self.s.Project.from_xml(model_xml)
        else:
            proj1 = project

        proj1_xml = tm.normalize_xml(proj1.to_xml(pretty=True))

        proj2 = self.s.load_project(model_file, start=False)
        proj2_xml = tm.normalize_xml(proj2.to_xml(pretty=True))

        tm.elements_equal(model_xml, proj1_xml)
        tm.elements_equal(model_xml, proj2_xml)

    def _test_copy(self, model, contquery, window):
        proj1 = self._load_project(model)

        # Shallow copy
        xml1 = proj1[contquery][window].to_xml(pretty=True)
        xml2 = proj1[contquery][window].copy().to_xml(pretty=True)

        self.assertEqual(xml1, xml2)

        # Deep copy
        xml2 = proj1[contquery][window].copy(deep=True).to_xml(pretty=True)

        self.assertEqual(xml1, xml2)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import Window, attribute
from .features import SchemaFeature
from .utils import get_args, ensure_element, connectors_to_end
from ..utils import xml


class FieldPlugin(object):
    '''
    Aggregate field plugin

    Parameters
    ----------
    plugin : string
        Name of the shared library.
    function : string
        Name of the function in the shared library.
    additive : bool, optional
        Specify for Aggregate windows only. Defaults to false.
    additive_insert_only : bool, optional
        Specify for Aggregate windows only. Defaults to false.

    Returns
    -------
    :class:`FieldPlugin`

    '''

    def __init__(self, plugin, function, additive=False, additive_insert_only=False):
        self.plugin = plugin
        self.function = function
        self.additive = additive
        self.additive_insert_only = additive_insert_only

    def copy(self, deep=False):
        return type(self)(self.plugin, self.function, additive=self.additive,
                          additive_insert_only=self.additive_insert_only)


class AggregateWindow(Window, SchemaFeature):
    '''
    Aggregate window

    Parameters
    ----------
    name : string
        The name of the window
    schema : Schema
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for the
        window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single
        update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval.
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window.
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    pubsub_index_type : int, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.

    Attributes
    ----------
    field_expressions : list-of-FieldExpressions
        Specifies Expression Engine Language (EEL) expressions assigned
        to a field.
    field_plugins : list-of-FieldPlugins
        Functions in a shared library whose returned value is assigned
        to a field.

    Returns
    -------
    :class:`AggregateWindow`

    '''

    window_type = 'aggregate'

    def __init__(self, name=None, schema=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_max_string=None, index_type=None,
                 pubsub_index_type=None):
        Window.__init__(self, **get_args(locals()))
        self.field_expressions = []
        self.field_plugins = []

    def copy(self, deep=False):
        out = Window.copy(self, deep=deep)
        out.field_expressions = list(self.field_expressions)
        if deep:
            out.field_plugins = []
            for item in self.field_plugins:
                out.field_plugins.append(item.copy(deep=deep))
        else:
            out.field_plugins = list(self.field_plugins)
        return out

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`AggregateWindow`

        '''
        out = super(AggregateWindow, cls).from_element(data, session=session)

        for item in data.findall('./output/field-expr'):
            out.field_expressions.append(item.text)
        for item in data.findall('./output/field-plug'):
            attrs = item.attrib
            out.field_plugins.append(
                FieldPlugin(attrs['plugin'],
                            attrs['function'],
                            additive=attrs.get('additive', 'f').startswith('t'),
                            additive_insert_only=attrs.get('additive_insert_only',
                                                           'f').startswith('t')))
        return out

    from_xml = from_element

    def to_element(self, query=None):
        '''
        Convert object to Element

        Parameters
        ----------
        query : string, optional
            Name of the continuous query

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = Window.to_element(self, query=None)

        if self.field_expressions or self.field_plugins:
            output = xml.add_elem(out, 'output')

            for item in self.field_expressions:
                xml.add_elem(output, 'field-expr', text_content=item)

            for item in self.field_plugins:
                xml.add_elem(output, 'field-plug',
                             attrib=dict(plugin=item.plugin,
                                         function=item.function,
                                         additive=item.additive,
                                         additive_insert_only=item.additive_insert_only))

        connectors_to_end(out)

        return out

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def add_field_expressions(self, *expr):
        '''
        Add aggregate field expression

        Parameters
        ----------
        expr : string
            The aggregation expression

        '''
        for item in expr:
            self.field_expressions.append(item)

    add_field_expr = add_field_expressions

    add_field_exprs = add_field_expressions

    add_field_expression = add_field_expressions

    def add_field_plugin(self, plugin, function, additive=False,
                         additive_insert_only=False):
        '''
        Add aggregate field plugin

        Parameters
        ----------
        plugin : string
            Name of the shared library.
        function : string
            Name of the function in the shared library.
        additive : bool, optional
            Specify for Aggregate windows only. Defaults to false.
        additive_insert_only : bool, optional
            Specify for Aggregate windows only. Defaults to false.

        '''
        self.field_plugins.append(FieldPlugin(plugin, function, additive=additive,
                                              additive_insert_only=additive_insert_only))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Bacnet Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class BacnetPublisher(Connector):
    '''
    Publish Bacnet events

    Parameters
    ----------
    bacnetbbmdaddress : string
        Specifies the IP address of the BBMD
    bacnetbbmdport : int
        Specifies the port of the BBMD
    bacnetconfigfile : string
        Specifies the JSON configuration file containing
        Bacnet device and object
    bacnetipport : int, optional
        Specifies the local port used by the connector.
        The default port number is 47808.
    blocksize : int, optional
        Specifies the number of events to include in a published
        event block. The default value is 1.
    configfilesection : string, optional
        Specifies the name of the section in the connector config
        file to parse for configuration parameters. Specify the
        value as [configfilesection].
    ignoretimeouts : string, optional
        Logs a warning and continues if an attempt to read a
        property from a Bacnet device results in a timeout.
        The default is to log an error and stop.
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert.
    maxevents : int, optional
        Specifies the maximum number of events to publish.
    transactional : string, optional
        Sets the event block type to transactional. The default
        value is normal.

    Returns
    -------
    :class:`BacnetPublisher`

    '''
    connector_key = dict(cls='bacnet', type='publish')

    property_defs = dict(
        bacnetbbmdaddress=prop('bacnetbbmdaddress', dtype='string', required=True),
        bacnetbbmdport=prop('bacnetbbmdport', dtype='int', required=True),
        bacnetconfigfile=prop('bacnetconfigfile', dtype='string', required=True),
        bacnetipport=prop('bacnetipport', dtype='int'),
        blocksize=prop('blocksize', dtype='int'),
        configfilesection=prop('configfilesection', dtype='string'),
        ignoretimeouts=prop('ignoretimeouts', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int'),
        transactional=prop('transactional', dtype='string')
    )

    def __init__(self, bacnetbbmdaddress=None, bacnetbbmdport=None,
                 bacnetconfigfile=None, name=None, is_active=None,
                 bacnetipport=None, blocksize=None, configfilesection=None,
                 ignoretimeouts=None, publishwithupsert=None,
                 maxevents=None, transactional=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'bacnet', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['bacnetbbmdaddress',
                                                   'bacnetbbmdport',
                                                   'bacnetconfigfile'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Logger '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
from six.moves import urllib
from .base import ESPObject
from .utils.rest import get_params
from .utils.data import get_project_data


class Logger(ESPObject):
    '''
    Server logger

    Attributes
    ----------
    name : string
        Name of the logger
    level : string
        Log level

    Parameters
    ----------
    name : string
        Name of the logger
    level : string
        Log level.  This shouuld be one of the following class variables:
            * Logger.INFO
            * Logger.TRACE
            * Logger.ERROR
            * Logger.WARN
            * Logger.FATAL
            * Logger.DEBUG
            * Logger.OFF

    '''

    INFO = 'info'
    TRACE = 'trace'
    ERROR = 'error'
    WARN = 'warn'
    WARNING = 'warn'
    FATAL = 'fatal'
    DEBUG = 'debug'
    OFF = 'off'

    def __init__(self, name=None, level=None):
        ESPObject.__init__(self)
        self.name = name
        if level is not None:
            level = level.lower()
            if level == 'warning':
                level = 'warn'
        self._level = level

    def __str__(self):
        return "%s(name='%s', level='%s')" % (type(self).__name__, self.name, self.level)

    def __repr__(self):
        return str(self)

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        self.set_level(value)

    def set_level(self, level):
        '''
        Set the log level

        Parameters
        ----------
        level : string
            Log level.  This should be one of the following:
                * Logger.INFO
                * Logger.TRACE
                * Logger.ERROR
                * Logger.WARN
                * Logger.FATAL
                * Logger.DEBUG
                * Logger.OFF

        '''
        level = level.lower()
        if level == 'warning':
            level = 'warn'
        levels = ['info', 'trace', 'error', 'warn', 'fatal', 'debug', 'off']
        if level not in levels:
            raise ValueError('%s is not a valid logging level' % level)
        self._put('loggers/%s/level' % self.name, params=get_params(value=level))
        self._level = level.lower()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Initialization of ESP options

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import functools
import pandas as pd
from .utils.config import (register_option, check_boolean, check_int, get_option,
                           set_option, reset_option, describe_option, check_url,
                           ESPOptionError, check_string, options, get_suboptions,
                           get_default, check_float, option_context)


# Root of server URLs
ESP_ROOT = 'SASESP'

# Store pandas version
PANDAS_VERSION = tuple([int(x) for x in pd.__version__.split('.')])

CONCAT_OPTIONS = {}
if PANDAS_VERSION >= (0, 23, 0):
    CONCAT_OPTIONS['sort'] = True

#
# Connection options
#

register_option('hostname', 'string', check_string,
                'localhost',
                'Specifies the hostname for the ESP server.',
                environ='ESPHOST')

register_option('port', 'int', check_int, 0,
                'Sets the port number for the ESP server.',
                environ='ESPPORT')

register_option('protocol', 'string', check_string, 'http',
                'Sets the protocol of the ESP server.',
                environ='ESPPROTOCOL')

#
# Project options
#

register_option('display.show_schema', 'boolean', check_boolean, False,
                'Should schemas be displayed when rendering project diagrams?')

register_option('display.show_field_type', 'boolean', check_boolean, True,
                'Should field types be displayed when rendering project diagrams?')

register_option('display.image_scale', 'float',
                functools.partial(check_float, minimum=0.1, maximum=10), 1.0,
                'Specifies the scale factor for rendering project diagrams.')

register_option('display.max_fields', 'int',
                functools.partial(check_int, minimum=0), 20,
                'Specifies the maximum number of schema fields per window to display\n'
                'when rendering a project diagram.')

#
# Debug options
#

register_option('debug.requests', 'boolean', check_boolean, False,
                'Display requested URL when accessing REST interface.')
register_option('debug.request_bodies', 'boolean', check_boolean, False,
                'Display body of request when accessing REST interface.')
register_option('debug.responses', 'boolean', check_boolean, False,
                'Display raw responses from server.')
register_option('debug.events', 'boolean', check_boolean, False,
                'Display raw events from server.')



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Utility for handling message resources

'''

from rbtranslations import Translations
import rbtranslations

import os

class Resources(object):
    _instance = None

    @staticmethod
    def getInstance():
        if Resources._instance == None:
            Resources._instance = Resources()

        return(Resources._instance)

    def __init__(self):
        self._bundles = {}

        path = os.path.abspath (os.path.join(os.path.dirname(__file__), "../locales/pymsg.properties"))
        #with open(path,mode="r",encoding="utf-8") as fp:
        #with open(path,encoding="iso-8859-1") as fp:
        #with open(path,mode="rb") as fp:
        with open(path,mode="r") as fp:
            self._default = Translations(fp)

    def getText(self,key,locale = None):
        text = None
        bundle = None

        if locale != None:
            if locale in self._bundles:
                bundle = self._bundles[locale]
        else:
            bundle = self._default

        if bundle != None:
            text = bundle.gettext(key)

        return(text)



try:
    import graphviz as gv
except ImportError:
    raise ImportError("The graphviz module is required for exporting to graphs.")

from ..utils.notebook import scale_svg
import ipywidgets as widgets

import logging
import re

import esppy.espapi.tools as tools

from esppy.espapi.tools import Options

class ViewerBase(widgets.VBox,Options):
    def __init__(self,visuals,connection,**kwargs):
        widgets.VBox.__init__(self,layout=widgets.Layout(border=visuals.getOpt("border","0"),padding=visuals.getOpt("padding","10px"),margin=visuals.getOpt("margin","10px")))
        Options.__init__(self,**kwargs)
        self._visuals = visuals
        self._connection = connection

    def setWidth(self,value):
        self.setOpt("width",value)

    def setHeight(self,value):
        self.setOpt("height",value)

class ModelViewer(ViewerBase):
    def __init__(self,visuals,connection,**kwargs):
        ViewerBase.__init__(self,visuals,connection,**kwargs)
        self._stats = self._connection.getStats()
        self._data = None
        self._project = "*"
        self._model = None
        self._projects = None
        self._memory = None

        self._gradient = tools.Gradient("#ffffff",levels=100,min=0,max=100)

        self._windowColors = None

        colors = self._visuals._colors.getSpread(5)
        self._windowColors = {}
        for i,type in enumerate(["input","utility","analytics","textanalytics","transformation"]):
            self._windowColors[type] = colors[i]

        width = self.getOpt("width","98%",True)
        height = self.getOpt("height","400px",True)

        self._projects = widgets.Dropdown(description="Projects")

        self._projects.observe(self.setProject,names="value")

        w = []

        l1 = layout=widgets.Layout(margin="0px 0px 0px 10px")
        l2 = layout=widgets.Layout(margin="0px 10px 0px 5px")

        self._systemMem = widgets.Label(value="0",layout=l2)
        self._virtualMem = widgets.Label(value="0",layout=l2)
        self._residentMem = widgets.Label(value="0",layout=l2)

        w.append(widgets.Label(value="System Memory:",layout=l1))
        w.append(self._systemMem)
        w.append(widgets.Label(value="Virtual Memory:",layout=l1))
        w.append(self._virtualMem)
        w.append(widgets.Label(value="Resident Memory:",layout=l1))
        w.append(self._residentMem)

        cbLayout = widgets.Layout(max_width="100px")

        self._memory = widgets.HBox(w)

        showcpu = widgets.Checkbox(description="CPU",indent=False,layout=widgets.Layout(max_width="100px",margin="0px 0px 0px 50px"),value=self.getOpt("cpu"))
        showcounts = widgets.Checkbox(description="Counts",indent=False,layout=cbLayout,value=self.getOpt("counts"))
        showtypes = widgets.Checkbox(description="Types",indent=False,layout=cbLayout,value=self.getOpt("type"))
        showindices = widgets.Checkbox(description="Indices",indent=False,layout=cbLayout,value=self.getOpt("index"))
        showschema = widgets.Checkbox(description="Schema",indent=False,layout=cbLayout,value=self.getOpt("schema"))

        show = widgets.HBox([showcpu,showcounts,showtypes,showindices,showschema])

        box = widgets.HBox([self._projects,show])
        #top = widgets.VBox([box,self._memory],layout=widgets.Layout(border="1px solid #c0c0c0"))
        top = widgets.VBox([box,self._memory],layout=widgets.Layout(border="1px solid #c0c0c0",padding="10px",margin="5px 5px 0px 5px"))

        #self._graph = widgets.HTML(layout=widgets.Layout(border="1px solid #c0c0c0",overflow="auto",padding="10px"))
        self._graph = widgets.HTML(layout=widgets.Layout(border="1px solid #c0c0c0",overflow="auto",margin="5px"))
        #self._html = widgets.VBox([top,self._graph],layout=widgets.Layout(border="1px solid #c0c0c0"))
        #self._html = widgets.VBox([top,self._graph],layout=widgets.Layout(border="1px solid #c0c0c0"))

        showcpu.observe(self.showCpu,names="value")
        showcounts.observe(self.showCounts,names="value")
        showtypes.observe(self.showTypes,names="value")
        showindices.observe(self.showIndices,names="value")
        showschema.observe(self.showSchema,names="value")

        self.setStats()

        self._connection.loadModel(self)

        self.children = [top,self._graph]

    def setStats(self):
        cpu = self.getOpt("cpu",False)
        counts = self.getOpt("counts",False)
        cpuColor = self.getOpt("cpu_color")

        if cpu or counts or (cpuColor != None):
            if self._data == None:
                self._data = {}
                self._stats.addDelegate(self)
            self._stats.setOpt("counts",counts)
        else:
            self._data = None
            self._stats.removeDelegate(self)

    def showCpu(self,b):
        self.setOpt("cpu",b["new"])
        self.setStats()
        self.setContent()

    def showCounts(self,b):
        self.setOpt("counts",b["new"])
        self.setStats()
        self.setContent()

    def showTypes(self,b):
        self.setOpt("type",b["new"])
        self.setContent()

    def showIndices(self,b):
        self.setOpt("index",b["new"])
        self.setContent()

    def showSchema(self,b):
        self.setOpt("schema",b["new"])
        self.setContent()

    def setContent(self):
        content = self.build()
        if content != None:
            self._graph.value = content

    def modelLoaded(self,model,conn):
        self._model = model
        self.setContent()

    def handleStats(self,stats):
        self._data = {}
        data = stats.getData();
        memory = stats.getMemoryData();

        for o in data:
            key = o["project"]
            key += "/"
            key += o["contquery"]
            key += "/"
            key += o["window"]
            self._data[key] = o;

        if memory != None:
            self._systemMem.value = str(memory["system"])
            self._virtualMem.value = str(memory["virtual"])
            self._residentMem.value = str(memory["resident"])

        self.setContent()

    def build(self):
        if self._model == None:
            return(None)

        a = [("ALL","*")]

        if self._projects != None:
            for p in self._model._projects:
                a.append((p["name"],p["name"]))

            self._projects.options = a

            if self._project != "*":
                self._projects.value = self._project

        rankdir = "LR"

        opt = self.getOpt("orientation","horizontal")

        if opt == "vertical":
            rankdir = "TB"

        graphAttr = {}
        graphAttr["style"] = "filled,rounded"
        graphAttr["color"] = "#58a0d3"
        graphAttr["fillcolor"] = "#f8f8f8"
        graphAttr["rankdir"] = rankdir
        graphAttr["center"] = "true"

        nodeAttr = {}
        nodeAttr["shape"] = "rect"
        nodeAttr["fontname"] = "helvetica"
        nodeAttr["fontsize"] = "12"
        nodeAttr["style"] = "filled,bold"
        nodeAttr["color"] = "#58a0d3"
        nodeAttr["fillcolor"] = "#c8f0ff"

        edgeAttr = {}
        edgeAttr["splines"] = "spline"

        graphs = {}

        container = None

        if self._project == "*":

            container = gv.Digraph(graph_attr=graphAttr,node_attr=nodeAttr,edge_attr=edgeAttr,format="svg")

            for i,p in enumerate(self._model._projects):
                graph = gv.Digraph(name="cluster_" + str(i),graph_attr=graphAttr,node_attr=nodeAttr,edge_attr=edgeAttr,format="svg")
                graph.attr(label=p["name"],labeljust='l')
                graphs[p["name"]] = graph
        else:
            graph = gv.Digraph(name="cluster_1",graph_attr=graphAttr,node_attr=nodeAttr,edge_attr=edgeAttr,format="svg")
            #graph.attr(label=self._project,labeljust='l')
            graphs[self._project] = graph

        cpuColor = self.getOpt("cpu_color")

        if cpuColor != None:
            self._gradient.color = tools.Colors.getColorFromName(cpuColor)

        showCpu = self.getOpt("cpu",False)
        showCounts = self.getOpt("counts",False)
        showType = self.getOpt("type",False)
        showIndex = self.getOpt("index",False)
        showSchema = self.getOpt("schema",False)

        if self._memory != None:
            if showCpu:
                self._memory.layout.display = "flex"
            else:
                self._memory.layout.display = "none"

        showProperties = showCpu or showCounts or showType or showIndex

        for a in self._model._windows:
            if self._project == "*" or a["p"] == self._project:
                if a["type"] == "source" or a["type"] == "window-source" or len(a["incoming"]) > 0 or len(a["outgoing"]) > 0:
                    label = []
                    label.append("<<table border='0' cellspacing='0' cellpadding='0'>")
                    label.append("<tr><td>" + a["name"] + "</td></tr>")
                    if showProperties:
                        label.append("<tr><td><table border='0' cellspacing='0' cellpadding='1'>")
                    if showType:
                        label.append("<tr><td align='right'>type:</td><td> </td><td align='left'>" + a["type"] + "</td></tr>")
                    color = None
                    cpu = None
                    count = None
                    if self._data != None:
                        o = self._data.get(a["key"])
                        cpu = 0
                        count = 0
                        if o != None:
                            cpu = int(o["cpu"])
                            count = int(o["count"])

                        if showCpu:
                            label.append("<tr><td align='right'>cpu:</td><td> </td><td align='left'>" + "{0:5}".format(cpu) + "</td></tr>")
                        if cpuColor:
                            color = self._gradient.darken(cpu)
                        if showCounts:
                            label.append("<tr><td align='right'>count:</td><td> </td><td align='left'>" + "{0:5}".format(count) + "</td></tr>")
                    if showIndex:
                        label.append("<tr><td align='right'>index:</td><td> </td><td align='left'>" + a["index"] + "</td></tr>")
                    if showProperties:
                        label.append("</table></td></tr>")
                    if showSchema:
                        label.append("<tr><td>")
                        label.append("<table border='0' cellspacing='0' cellpadding='1'>")
                        label.append("<tr><td colspan='2'> </td></tr>")
                        label.append("<tr><td colspan='2'>Schema</td></tr>")
                        for f in a["schema"]._fields:
                            if f["isKey"]:
                                label.append("<tr><td align='left'>" + f["name"] + "*</td><td align='left'>" + f["type"] + "</td></tr>")
                        for f in a["schema"]._fields:
                            if f["isKey"] == False:
                                label.append("<tr><td align='left'>" + f["name"] + "</td><td align='left'>" + f["type"] + "</td></tr>")
                        label.append("</table>")
                        label.append("</td></tr>")
                    label.append("</table>>")
                    label = "".join(label)
                    if color == None:
                        if self._windowColors != None:
                            color = self._windowColors.get(a["class"])

                    graph = graphs[a["p"]]

                    if color != None:
                        graph.node(a["key"],label,fillcolor=color)
                    else:
                        graph.node(a["key"],label)
                        
                    for z in a["outgoing"]:
                        graph.edge(a["key"],z["key"])

        if container != None:
            for key,value in graphs.items():
                container.subgraph(value)
            return(container._repr_svg_())

        return(graph._repr_svg_())

    def setProject(self,change):
        self.project = self._projects.value

    def dataChanged(self,datasource,data,clear):
        pass

    @property
    def project(self):
        return(self._project)

    @project.setter
    def project(self,value):
        self._project = value;
        self.setContent()

class LogViewer(ViewerBase):
    def __init__(self,visuals,connection,**kwargs):
        ViewerBase.__init__(self,visuals,connection,**kwargs)

        width = self.getOpt("width","98%")
        height = self.getOpt("height","200px")

        self._max = self.getOpt("max",50);

        self._bg = self.getOpt("bg","#f8f8f8")
        self._border = self.getOpt("border","1px solid #d8d8d8")

        components = []
        self._log = widgets.HTML(value="",layout=widgets.Layout(width=width,height=height,border=self._border,overflow="auto"))
        components.append(self._log)

        self._filter = self.getOpt("filter")
        self._regex = None

        if self._filter != None:
            self._filterText = widgets.Text(description="Filter",value=self._filter,layout=widgets.Layout(width="70%"))
            if len(self._filter) > 0:
                self._regex = re.compile(self._filter,re.I)
            setButton = widgets.Button(description="Set")
            clearButton = widgets.Button(description="Clear")
            setButton.on_click(self.filter)
            clearButton.on_click(self.clearFilter)
            components.append(widgets.HBox([self._filterText,setButton,clearButton]))

        self._box = widgets.VBox(components,layout=widgets.Layout(width="100%"))

        s = ""
        s += "<div style='width:100%;height:100%;background:" + self._bg + "'>"
        s += "</div>"
        self._log.value = s

        self._messages = []

        self._connection.getLog().addDelegate(self)
        
        self.children = [self._box]

    def handleLog(self,connection,message):
        self._messages.append(message)

        if self._max != None and len(self._messages) > self._max:
            diff = len(self._messages) - self._max

            for i in range(0,diff):
                self._messages.pop(self._max + i)

        self.load()

    def load(self):
        s = ""

        s += "<div style='width:100%;height:100%;background:" + self._bg + "'>"
        s += "<pre style='width:100%;height:100%;background:" + self._bg + "'>"

        #for message in self._messages:
        for message in reversed(self._messages):
            if self._regex != None:
                if self._regex.search(message) == None:
                    continue
                
            s += message;
            s += "\n"
            s += "\n"

        s += "</pre>"
        s += "</div>"

        self._log.value = s

    def filter(self,b):
        self._filter = self._filterText.value.strip()
        if len(self._filter) == 0:
            self._filter = None
        else:
            self._regex = re.compile(self._filter,re.I)
        self.load()

    def clearFilter(self,b):
        self._filterText.value = ""
        self._filter = None
        self._regex = None
        self.load()

class StatsViewer(ViewerBase):
    def __init__(self,visuals,connection,**kwargs):
        ViewerBase.__init__(self,visuals,connection,**kwargs)
        self._stats = self._connection.getStats();
        self._stats.addDelegate(self)

        width = self.getOpt("width","90%",True)
        height = self.getOpt("height","400px",True)

        self._stats.setOpts(**self.options)

        self._log = widgets.HTML()

        self.children = [self._log]

    def handleStats(self,stats):
        data = stats.getData()
        s = ""
        s += "<table border='1' cellspacing='0' cellpadding='4'>"
        s += "<tr>"
        s += "<td>Project</td>"
        s += "<td>Contquery</td>"
        s += "<td>Window</td>"
        s += "<td>CPU</td>"
        s += "</tr>"

        for o in data:
            if o["cpu"] >= 1:
                s += "<tr>"
                s += "<td>" + o["project"] + "</td>"
                s += "<td>" + o["contquery"] + "</td>"
                s += "<td>" + o["window"] + "</td>"
                s += "<td>" + str(int(o["cpu"])) + "</td>"
                s += "</tr>"

        s += "</table>"

        self._log.value = s



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import xml.etree.ElementTree as ET
import esppy

class MasStores(object):
    def __init__(self,esppy):
        self._url = esppy.url
        self._session = esppy.session
        self._stores = {}
        self.load()

    def load(self):
        response = self._session.get(self._url + "masStores")

        xml = ET.fromstring(response.text)

        stores = xml.findall(".//mas-store")

        names = []

        for s in stores:
            name = s.attrib["name"]
            names.append(name)

            if (name in self._stores) == False:
                self._stores[name] = MasStore(name,self)

            self._stores[name].set(s)

        remove = []

        for store in self._stores.keys():
            if (store in names) == False:
                remove.append(store)

        for store in remove:
            del self._stores[store]

    def getStores(self):
        return(self._stores)

    def getStore(self,name):
        store = None
        if name in self._stores:
            store = self._stores[name]
        return(store)

    def createStore(self,name):
        if name in self._stores:
            raise Exception("Mas store " + name + " already exists")

        response = self._session.put(self._url + "masStores/" + name)

        store = MasStore(name,self)

        self._stores[name] = store

        return(store)

    def __str__(self):
        s = ""
        for store in self._stores.values():
            s += "\t" + str(store) + "\n"
        return(s)

class MasStore(object):
    def __init__(self,name,stores):
        self._name = name
        self._stores = stores
        self._versions = {}

    def set(self,xml):
        versions = xml.findall(".//version")

        for v in versions:
            number = v.attrib["number"]
            version = MasStoreVersion(number,self)
            version.set(v)
            self._versions[number] = version

    def getObject(self,name,version="1.0"):
        o = None
        if version in self._versions:
            v = self._versions[version]
            o = v.getObject(name)
        return(o)

    def setObject(self,name,code,version="1.0"):
        url = self._stores._url + "masStoreObjects/" + self._name + "/" + name + "/" + version
        response = self._stores._session.put(url,data=code.encode("utf-8"))
        if response.status_code == 200:
            self._stores.load()
        return(response)

    def setObjectFromFile(self,name,path,version="1.0"):
        response = None
        with open(path) as f:
            contents = f.read()
            response = self.setObject(name,contents,version)
        return(response)

    def delete(self):
        url = self._stores._url + "masStores/" + self._name
        print("URL: " + url)
        response = self._stores._session.delete(url)
        self._stores.load()
        return(response.status_code)

    def __str__(self):
        s = ""
        s += self._name
        s += "\n"
        for version in self._versions.values():
            s += "\t" + str(version) + "\n"
        return(s)

class MasStoreVersion(object):
    def __init__(self,number,store):
        self._number = number
        self._store = store
        self._objects = {}

    def getObject(self,name):
        o = None

        if name in self._objects:
            o = self._objects[name]

        return(o)

    def set(self,xml):
        objects = xml.findall(".//mas-object")

        for o in objects:
            name = o.attrib["name"]
            obj = MasStoreObject(name,self)
            self._objects[name] = obj

    def __str__(self):
        s = ""
        s += self._number
        s += "\n"
        for o in self._objects.values():
            s += "\t" + str(o) + "\n"
        return(s)

class MasStoreObject(object):
    def __init__(self,name,version):
        self._name = name
        self._version = version
  
    @property
    def code(self):
        url = self._version._store._stores._url + "masStoreObjects/" + self._version._store._name + "/" + self._name + "/" + self._version._number
        response = self._version._store._stores._session.get(url)
        content = None
        if response.status_code == 200:
            content = response.text
        return(content)

    def delete(self):
        url = self._version._store._stores._url + "masStoreObjects/" + self._version._store._name + "/" + self._name + "/" + self._version._number
        response = self._version._store._stores._session.delete(url)
        self._version._store._stores.load()
        return(response.status_code)

    @property
    def name(self):
        return(self._name)

    @property
    def version(self):
        return(self._version)

    def __str__(self):
        s = ""
        s += "\t" + self._name + "\n"
        return(s)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestComputeWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('compute_field_expression.xml')

    def test_field_expression(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('name:string','ID*:int32','city:string'))
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='compute_field_expression.csv',
                          transactional=True,
                          blocksize='1')

        compute_win = esp.ComputeWindow(name='computeWindow',
                                        schema=('ID*:int32', 'name:string',
                                                'city:string','match:string'))
        compute_win.set_expr_initializer(expr='integer counter counter=0',type='int32')
        compute_win.add_field_expressions('name', 'city',
                                          'counter=counter+1 return counter')
        compute_win.add_connector('fs',
                                   conn_name='sub', conn_type='subscribe',
                                   type='sub',
                                   fstype='csv',
                                   fsname='result.out',
                                   snapshot=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(compute_win)

        src.add_target(compute_win)

        self._test_model_file('compute_field_expression.xml', proj)

    @unittest.skip('Need way to load plugin')
    def test_plugin(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('ID*:int64','symbol:string',
                                       'price:double','markup:double'),
                               index_type='hash', collapse_updates=True)
        src.add_connector('fs',
                          conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='compute_plugin.csv',
                          transactional=True,
                          blocksize='1')

        compute_win = esp.ComputeWindow(name='compute',
                                        schema=('ID*:int64','symbol:string',
                                                'price:double','markup:double',
                                                'sum:double','diff:double',
                                                'prod:double','quot:double'),
                                        collapse_updates=True)
        compute_win.set_context_plugin('libmethod','get_context')
        compute_win.add_field_plugin('libmethod', ['copyScalarFunction',
                                                   'copyScalarFunction',
                                                   'copyScalarFunction',
                                                   'arithScalarFunction',
                                                   'arithScalarFunction',
                                                   'arithScalarFunction',
                                                   'arithScalarFunction'])
        compute_win.add_connector('fs',
                                  conn_name='sub', conn_type='subscribe',
                                  type='sub',
                                  fstype='csv',
                                  fsname='result.out',
                                  snapshot=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(compute_win)

        src.add_target(compute_win)

        self._test_model_file('compute_plugin.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import datetime
import os
import six
import esppy
import sys
import unittest
from . import utils as tm
from esppy.router import Router

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestRouter(tm.TestCase):

    pub_xml = r'''<esp-router name="router">
  <esp-engines>
    <esp-engine host="localhost" name="esp1" port="42002" />
    <esp-engine host="localhost" name="esp2" port="52002" />
  </esp-engines>
  <esp-destinations>
    <publish-destination name="src" opcode="upsert">
      <publish-target>
        <engine-func>esp2</engine-func>
        <project-func>proj2</project-func>
        <contquery-func>VERT_PEAK_FORCE</contquery-func>
        <window-func>READINGS_INPUT</window-func>
      </publish-target>
    </publish-destination>
  </esp-destinations>
  <esp-routes>
    <esp-route name="source" snapshot="false" to="src">
      <engine-expr>esp1</engine-expr>
      <project-expr>proj1</project-expr>
      <contquery-expr>VERT_PEAK_FORCE</contquery-expr>
      <window-expr>READINGS_INPUT</window-expr>
    </esp-route>
  </esp-routes>
</esp-router>'''

    write_xml = r'''<esp-router name="router">
  <esp-engines>
    <esp-engine host="localhost" name="esp1" port="42002" />
  </esp-engines>
  <esp-destinations>
    <writer-destination dateformat="%Y%m%dT%H:%M:%S.%f" format="csv" name="out1">
      <file-func>string('./','output1.csv')</file-func>
    </writer-destination>
    <writer-destination dateformat="%Y%m%dT%H:%M:%S.%f" format="csv" name="out2">
      <file-func>string('./','output2.csv')</file-func>
    </writer-destination>
    <publish-destination name="src" opcode="upsert">
      <publish-target>
        <engine-func>esp1</engine-func>
        <project-func>proj2</project-func>
        <contquery-func>VERT_PEAK_FORCE</contquery-func>
        <window-func>READINGS_INPUT</window-func>
      </publish-target>
    </publish-destination>
  </esp-destinations>
  <esp-routes>
    <esp-route name="output1" snapshot="false" to="out1">
      <engine-expr>esp1</engine-expr>
      <project-expr>proj1</project-expr>
      <contquery-expr>VERT_PEAK_FORCE</contquery-expr>
      <window-expr>READINGS_INPUT</window-expr>
    </esp-route>
    <esp-route name="output2" snapshot="false" to="out2">
      <engine-expr>esp1</engine-expr>
      <project-expr>proj2</project-expr>
      <contquery-expr>VERT_PEAK_FORCE</contquery-expr>
      <window-expr>READINGS_INPUT</window-expr>
    </esp-route>
    <esp-route name="source" snapshot="false" to="src">
      <engine-expr>esp1</engine-expr>
      <project-expr>proj1</project-expr>
      <contquery-expr>VERT_PEAK_FORCE</contquery-expr>
      <window-expr>READINGS_INPUT</window-expr>
    </esp-route>
  </esp-routes>
</esp-router>'''

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

    def test_xml(self):
        router = Router.from_xml(self.pub_xml)
        out_xml = router.to_xml(pretty=True).strip()

        self.assertEqual(self.pub_xml, out_xml)

    def test_pub_api(self):
        router = self.s.create_router('router')

        router.add_engine('localhost', 52002, name='esp2')
        router.add_engine('localhost', 42002, name='esp1')

        router.add_publish_destination('esp2.proj2.VERT_PEAK_FORCE.READINGS_INPUT', name='src', opcode='upsert')

        router.add_route('esp1.proj1.VERT_PEAK_FORCE.READINGS_INPUT', 'src', name='source', snapshot=False)

        xml = router.to_xml(pretty=True).strip()

        self.assertEqual(self.pub_xml, xml)

    def test_write_api(self):
        router = self.s.create_router(name='router')

        router.add_engine('localhost', '42002', name='esp1')

        router.add_publish_destination('esp1.proj2.VERT_PEAK_FORCE.READINGS_INPUT', name='src', opcode='upsert')
        router.add_writer_destination("string('./','output1.csv')", 'csv', name='out1')
        router.add_writer_destination("string('./','output2.csv')", 'csv', name='out2')

        router.add_route('esp1.proj1.VERT_PEAK_FORCE.READINGS_INPUT', 'src', name='source')
        router.add_route('esp1.proj1.VERT_PEAK_FORCE.READINGS_INPUT', 'out1', name='output1')
        router.add_route('esp1.proj2.VERT_PEAK_FORCE.READINGS_INPUT', 'out2', name='output2')
    
        xml = router.to_xml(pretty=True).strip()

        self.maxDiff = None
        self.assertEqual(self.write_xml, xml)


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals

import json
import re
import requests
import logging
import sys
import xml.etree.ElementTree as ET
from six.moves import urllib
from ..config import get_option
from ..exceptions import ESPError

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def to_camel(val):
    '''
    Convert underscore-delimited names to camel-case

    Parameters
    ----------
    val : string
        The string to convert

    Returns
    -------
    string

    '''
    return re.sub(r'_([A-Za-z])', lambda x: x.group(1).upper(), val)


def to_underscore(val):
    '''
    Convert camel-case names to underscore-delimited

    Parameters
    ----------
    val : string
        The string to convert

    Returns
    -------
    string

    '''
    out = re.sub(r'([A-Z])', lambda x: '_' + x.group(1).lower(), val)
    if re.match(r'[A-Z]', val):
        return out[1:]
    return out


def _to_string(value, ESPObject):
    if value is True:
        return 'true'
    elif value is False:
        return 'false'
    elif isinstance(value, ESPObject):
        return value.name
    if isinstance(value, (list, set, tuple)):
        out = []
        for item in value:
            out.append(_to_string(item, ESPObject).replace(' ', '%20'))
        return '|'.join(out)
    else:
        return '%s' % value


def get_params(**kwargs):
    '''
    Convert keyword arguments into URL parameters

    This function does the following conversions:

        * Underscore-delimited names to camel-case
        * True / False values to 'true' / 'false'
        * ['a', 'b', 'c'] to 'a|b|c'
        * ESPObject to ESPObject.name

    Parameters
    ----------
    **kwargs : keyword arguments, optional
       Arbitrary Python arguments to convert to URL parameters

    '''
    try:
        ESPObject    # noqa: F821
    except:
        from ..base import ESPObject

    out = {}
    for key, value in kwargs.items():
        if value is not None:
            out[to_camel(key)] = _to_string(value, ESPObject)
    return out


class RESTHelpers(object):
    '''
    Helper methods for running REST requests against an ESP server

    Attributes
    ----------
    session : requests.Session
        The session object
    base_url : string
        The base server URL
    url : string
        The URL of the object

    '''

    def __init__(self, session=None):
        self.session = session

    def _insert_params(self, url, **kwargs):
        # The REST server is a bit particular about what can and
        # can't be URL encoded.  Handle that here.
        params = kwargs.pop('params', {})
        params = '&'.join(['%s=%s' % (key, ('%s' % value).replace(' ', '%20'))
                           for key, value in params.items()])
        if '?' in url:
            params = '&%s' % params
        else:
            params = '?%s' % params
        return (url + params, kwargs)

    def _verify_session(self):
        if self.session is None:
            raise ValueError('There is no session associated with this object.')

    @property
    def session(self):
        ''' The server session object '''
        return getattr(self, '_session', None)

    @session.setter
    def session(self, session):
        setattr(self, '_session', session)

    @property
    def base_url(self):
        ''' The base server URL '''
        self._verify_session()
        return self.session.base_url

    @property
    def url(self):
        ''' The URL of the object '''
        self._verify_session()
        return self.session.base_url

    def _get(self, path=None, format='xml', raw=False, **kwargs):
        '''
        GET the specified path

        Parameters
        ----------
        path : string, optional
            Absolute or relative URL
        format : string, optional
            Output format: 'xml' or 'json'
        raw : boolean, optional
            Should the raw content be returned rather than an Element?
        **kwargs : keyword parameters, optional
            URL parameters

        Returns
        -------
        dict

        '''
        self._verify_session()

        if path and (path.startswith('http:') or path.startswith('https:')):
            url = path
        elif path:
            url = urllib.parse.urljoin(self.url, path)
        else:
            url = self.url

        if format == 'json':
            try:
                out = self.session.get(url, **kwargs)
                out = out.json()
                return out
            except JSONDecodeError:
                sys.stderr.write('%s\n' % out.content.decode('utf-8'))
                raise

        url, kwargs = self._insert_params(url, **kwargs)

        if get_option('debug.requests'):
            sys.stderr.write('GET %s\n' % url)

        if get_option('debug.request_bodies') and kwargs.get('data'):
            sys.stderr.write('%s\n' % kwargs['data'])

        content = self._error_check(self.session.get(url,
                                                     **kwargs)).content.decode('utf-8')

        if get_option('debug.responses'):
            sys.stderr.write('%s\n' % content)

        if raw:
            return content

        return ET.fromstring(content)

    def _post(self, path=None, **kwargs):
        '''
        POST the specified path

        Parameters
        ----------
        path : string, optional
            Absolute or relative URL
        **kwargs : keyword parameters, optional
            URL parameters

        Returns
        -------
        dict

        '''
        self._verify_session()

        if path and (path.startswith('http:') or path.startswith('https:')):
            url = path
        elif path:
            url = urllib.parse.urljoin(self.url, path)
        else:
            url = self.url

        url, kwargs = self._insert_params(url, **kwargs)

        if get_option('debug.requests'):
            sys.stderr.write('POST %s\n' % url)

        content = self._error_check(self.session.post(url, **kwargs)).content.decode('utf-8')

        if get_option('debug.responses'):
            sys.stderr.write('%s\n' % content)

        return ET.fromstring(content)

    def _put(self, path=None, **kwargs):
        '''
        PUT the specified path

        Parameters
        ----------
        path : string, optional
            Absolute or relative URL
        **kwargs : keyword parameters, optional
            URL parameters

        Returns
        -------
        dict

        '''
        self._verify_session()

        if path and (path.startswith('http:') or path.startswith('https:')):
            url = path
        elif path:
            url = urllib.parse.urljoin(self.url, path)
        else:
            url = self.url

        url, kwargs = self._insert_params(url, **kwargs)

        if get_option('debug.requests'):
            sys.stderr.write('PUT %s\n' % url)

        content = self._error_check(self.session.put(url, **kwargs)).content.decode('utf-8')

        if get_option('debug.responses'):
            sys.stderr.write('%s\n' % content)

        return ET.fromstring(content)

    def _delete(self, path=None, **kwargs):
        '''
        DELETE the specified path

        Parameters
        ----------
        path : string, optional
            Absolute or relative URL
        **kwargs : keyword parameters, optional
            URL parameters

        Returns
        -------
        dict

        '''
        self._verify_session()

        if path and (path.startswith('http:') or path.startswith('https:')):
            url = path
        elif path:
            url = urllib.parse.urljoin(self.url, path)
        else:
            url = self.url

        url, kwargs = self._insert_params(url, **kwargs)

        if get_option('debug.requests'):
            sys.stderr.write('DELETE %s' % url)

        content = self._error_check(self.session.delete(url, **kwargs)).content.decode('utf-8')

        if get_option('debug.responses'):
            sys.stderr.write('%s\n' % content)

        return ET.fromstring(content)

    def _error_check(self, resp):
        '''
        Check the server response for errors

        Parameters
        ----------
        resp : dict
            Data from server

        Returns
        -------
        ``resp``

        '''

        error = None

        if resp.status_code >= 400:
            try:
                elem = ET.fromstring(resp.content)
            except ET.ParseError:
                sys.stderr.write('%s\n' % resp.content)
                raise

            msg = elem.find('./message/response/message')
            if msg is not None:
                error = msg.text
                if get_option('debug.responses'):
                    sys.stderr.write('%s\n' % resp.content)

            msg = elem.find('./message')
            if msg is not None:
                error = msg.text
                if get_option('debug.responses'):
                    sys.stderr.write('%s\n' % resp.content)

            if get_option('debug.responses'):
                sys.stderr.write('%s\n' % resp.content)

            if error != None:
                details = elem.findall('./details/detail')
                for detail in details:
                    error += "\n";
                    error += detail.text;
            else:
                error = resp.content

        if error != None:
            raise ESPError(error)

        return resp



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .utils import get_args
import inspect
from .base import BaseWindow, attribute, INDEX_TYPES
from .features import (ParametersFeature, SchemaFeature, InputMapFeature,
                       OutputMapFeature, MASMapFeature, ConnectorsFeature)
from .calculate import CalculateWindow
from .helpers import generators


class PythonHelper(BaseWindow, SchemaFeature,
                   ParametersFeature, InputMapFeature, OutputMapFeature,
                   MASMapFeature, ConnectorsFeature):
    '''
    Python Window

    Notes
    -----
    This class is basically a CalculateWindow with the algorithm specified
    to 'MAS'.

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for the
        window and false disables it.
    description : string, optional
        Description of the window
    algorithm : string, optional
        The name of the algorithm
    input_map : dict, optional
        Input mappings
    output_map : dict, optional
        Output mappings
    index_type : string, optional
        Index type for the window
    produces_only_inserts : bool, optional
        Set to true when you know that the procedural window always
        produces inserts
    **parameters : keyword arguments, optional
        The parameters to the algorithm

    Returns
    -------
    :class:`PythonHelper`

    '''

    window_type = 'calculate'

    is_hidden = True

    algorithm = attribute('algorithm', dtype='string')
    index_type = attribute('index', dtype='string', values=INDEX_TYPES)
    produces_only_inserts = attribute('produces-only-inserts', dtype='bool')

    def __init__(self, name=None, schema=None, pubsub=None,
                 description=None, index_type=None,
                 produces_only_inserts=None, input_map=None,
                 output_map=None, mas_info_list=None, **parameters):
        algorithm = 'MAS'
        BaseWindow.__init__(self, **get_args(locals()))
        self.mas_info_list = []

    def _register_to_project(self, project_handle):
        for mas_info in self.mas_info_list:
            module_name = mas_info['module_name']
            entry_func_name = mas_info['entry_func_name']
            source = mas_info['source']
            # If the module doesn't already exist
            if module_name not in [mas_obj.module for mas_obj in project_handle.mas_modules]:
                code = mas_info['code']
                if code is None:
                    raise ValueError('To create a new MAS module, code string or function list must be provided')
                python_module = project_handle.create_mas_module(language="python",
                                                                 module=module_name,
                                                                 func_names='place_holder',
                                                                 code=code)
                project_handle.mas_modules.append(python_module)

            self.add_mas_window_map(module=module_name,
                                    source=source,
                                    function=entry_func_name)

    def add_mas_info(self, module_name, entry_func_name, source, funcs=None, code_file=None, inits=None):
        '''
        Add the information needed to create a MAS module and add MAS window map.

        Notes
        -----
        If code is specified, funcs is ignored.

        Parameters
        ----------
        module_name : string
            Name of the MAS module to be created
        entry_func_name : string
            Name of the entry function in the MAS module.
            An entry functions is supposed to consumes the events sent from
            its source window and returns desired outputs. Any entry function
            must have a doc string to describe its outputs.
        source : string
            Name of the source window
        funcs : list of callable functions
            Functions included in MAS module
        code_file : string
            Path to the Python code file
        inits : dict
            Initialization of global variables if needed

        Returns
        -------
        :class:`Project`

        Examples
        --------
        Create a MAS module with multiple functions and global variables

        >>> win.add_mas_info('foo_module', 'foo_1', 'source_win'
                             funcs=[foo_1, foo_2],
                             inits=dict(gv_1=[], gv_2=0))

        Create a MAS module with a code file. The entry function 'foo_1'
        is defined in the code string.

        >>> win.add_mas_info('foo_module', 'foo_1', 'source_win'
                             code_file='path/to/code.py')

        '''
        if funcs is None and code_file is None:
            code = None
        elif code_file is None:
            code = ''
            # extract code string from function list
            for func in funcs:
                try:
                    code = code + inspect.getsource(func) + '\n'
                except NameError:
                    raise ValueError('{} not found.'.format(func.__name__))
        else:
            code = open(code_file, 'r').read()

        # extract code string from initialization list
        if inits is not None:
            inits_str = '\n'
            for key, value in inits.items():
                inits_str += (key + '=' + str(value) + '\n')
            code = inits_str + code

        mas_info = {'module_name': module_name,
                    'entry_func_name': entry_func_name, 'code': code, 'source': source}
        self.mas_info_list.append(mas_info)


class KerasHelper(PythonHelper):
    '''
    Keras Window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for the
        window and false disables it.
    description : string, optional
        Description of the window
    algorithm : string, optional
        The name of the algorithm
    input_map : dict, optional
        Input mappings
    output_map : dict, optional
        Output mappings
    index_type : string, optional
        Index type for the window
    produces_only_inserts : bool, optional
        Set to true when you know that the procedural window always
        produces inserts
    **parameters : keyword arguments, optional
        The parameters to the algorithm

    Returns
    -------
    :class:`KerasHelper`

    '''

    def __init__(self, name=None, schema=None, pubsub=None,
                 description=None, index_type=None,
                 produces_only_inserts=None, input_map=None,
                 output_map=None, **parameters):
        PythonHelper.__init__(self, **get_args(locals()))

    def add_model_info(self, model_name, model_file, source, input_name='input', output_name='output', output_class='False'):
        """
        Add the information of a Keras model

        Parameters
        -----------
        model_name : string
            User-specified name of the model
        model_file : string
            The path to hdf5 file that stores the model structure and parameters.
            ESP server should be able to find this file.
        source : string
            Name of the source window
        input_name : string
            Name of input array (features).
            This name should match the schema of the source window
        output_name : string
            Name of output (predictions).
        output_class : bool
            If True, the output is the predicted class. If False, the output is
            an array of pedicted probabilities of each class.
            Only applicable to classification models.

        """
        code_generator = generators.KS_generator(model_file, input_name, output_name, output_class)
        code = code_generator.gen_wrap_str()
        mas_info = {'module_name': model_name,
                    'entry_func_name': 'ks_score', 'code': code, 'source': source}
        self.mas_info_list.append(mas_info)


class TensorflowHelper(PythonHelper):
    '''
    Tensorflow Window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for the
        window and false disables it.
    description : string, optional
        Description of the window
    algorithm : string, optional
        The name of the algorithm
    input_map : dict, optional
        Input mappings
    output_map : dict, optional
        Output mappings
    index_type : string, optional
        Index type for the window
    produces_only_inserts : bool, optional
        Set to true when you know that the procedural window always
        produces inserts
    **parameters : keyword arguments, optional
        The parameters to the algorithm

    Returns
    -------
    :class:`TensorflowHelper`

    '''

    def __init__(self, name=None, schema=None, pubsub=None,
                 description=None, index_type=None,
                 produces_only_inserts=None, input_map=None,
                 output_map=None, **parameters):
        PythonHelper.__init__(self, **get_args(locals()))

    def add_model_info(self, model_name, model_file, input_op, score_op,
                       source, input_name='input', output_name='output'):
        """
        Add the information of a Tensorflow model

        Parameters
        -----------
        model_name : string
            User-specified name of the model
        model_file : string
            the path to meta file that stores the graph structure.
            The checkpoint files should be within the same directory with the meta file.
            ESP server should be able to find the files.
        input_op : string
            Name of input operation
        score_op : string
            Name of scoring operation
        source : string
            Name of the source window
        input_name : string
            Name of input array (features).
            This name should match the schema of the source window
        output_name : string
            Name of output (predictions).

        Notes
        -----
        The Tensorflow models are expoted in checkpoint files using tf.train.Saver class.
        The name of input and scoring operations should be specified when creating the model.
        """
        code_generator = generators.TF_generator(model_file, input_op, score_op,
                                                 input_name, output_name)
        code = code_generator.gen_wrap_str()
        mas_info = {'module_name': model_name,
                    'entry_func_name': 'tf_score', 'code': code, 'source': source}
        self.mas_info_list.append(mas_info)


class JMPHelper(PythonHelper):
    '''
    JMP Window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for the
        window and false disables it.
    description : string, optional
        Description of the window
    algorithm : string, optional
        The name of the algorithm
    input_map : dict, optional
        Input mappings
    output_map : dict, optional
        Output mappings
    index_type : string, optional
        Index type for the window
    produces_only_inserts : bool, optional
        Set to true when you know that the procedural window always
        produces inserts
    copy_vars : string or list of string, optional
        Add fields to the automatically generated schema
    **parameters : keyword arguments, optional
        The parameters to the algorithm

    Returns
    -------
    :class:`JMPHelper`

    Notes
    -----
    If no schema is provided users, a schema will be automatically generated
    based on the outputs of the score function from JMP.
    '''

    def __init__(self, name=None, schema=None, pubsub=None,
                 description=None, index_type=None,
                 produces_only_inserts=None, input_map=None,
                 output_map=None, copy_vars=None, **parameters):
        PythonHelper.__init__(self, **get_args(locals()))
        self.copy_vars = copy_vars

    def add_model_info(self, model_name, model_file, source):
        """
        Add the information of a JMP model

        Parameters
        -----------
        module_name : string
            Name of the MAS module to be created
        score_file : string
            The path to the Python file exported by JMP
        source : string
            Name of the source window
        """
        code_generator = generators.JMP_generator(model_file)
        code = code_generator.gen_wrap_str()
        mas_info = {'module_name': model_name,
                    'entry_func_name': 'jmp_score', 'code': code, 'source': source}
        self.mas_info_list.append(mas_info)

        if self._schema.schema_string == '':
            self.schema = code_generator._gen_schema(self.copy_vars)


setattr(CalculateWindow, PythonHelper.__name__, PythonHelper)
setattr(CalculateWindow, KerasHelper.__name__, KerasHelper)
setattr(CalculateWindow, TensorflowHelper.__name__, TensorflowHelper)
setattr(CalculateWindow, JMPHelper.__name__, JMPHelper)
